#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import threading
from dataclasses import dataclass
from typing import Optional

from .env import (
    NEPTUNE_API_TOKEN,
    NEPTUNE_PROJECT,
)

__all__ = (
    "Context",
    "get_context",
    "set_context",
    "set_project",
    "set_api_token",
    "validate_context",
)

from ..exceptions import (
    NeptuneApiTokenNotProvided,
    NeptuneProjectNotProvided,
)


@dataclass(frozen=True)
class Context:
    project: Optional[str] = None
    api_token: Optional[str] = None

    def with_project(self, project: str) -> "Context":
        """
        Copy the Context overwriting the project field.
        """
        if not project:
            raise ValueError("Project name must be provided")
        return Context(project=project, api_token=self.api_token)

    def with_api_token(self, api_token: str) -> "Context":
        """
        Copy the Context overwriting the api_token field.
        """
        if not api_token:
            raise ValueError("API token must be provided")
        return Context(project=self.project, api_token=api_token)


def set_project(project: str) -> Context:
    """
    Set the project in the context.
    Returns the set context.
    """
    global _context
    with _lock:
        _context = _context.with_project(project)
        return _context


def set_api_token(api_token: str) -> Context:
    """
    Set the API token in the context.
    Returns the set context.
    """
    global _context
    with _lock:
        _context = _context.with_api_token(api_token)
        return _context


def get_context() -> Context:
    """
    Return currently set global Context.
    """
    with _lock:
        return _context


def set_context(context: Optional[Context] = None) -> Context:
    """
    The context is automatically set from the environment variables (if they exist) on import of the module,
    but it's possible to override it with this function.

    If the argument is None, the global context is reset from environment variables.
    The following environment variables are used:
    - NEPTUNE_PROJECT
    - NEPTUNE_API_TOKEN

    Returns the set context.
    """
    global _context

    with _lock:
        _context = context or _context_from_env()
        return _context


def validate_context(context: Optional[Context] = None, validate_project: bool = False) -> Context:
    assert context is not None, "Context should have been set on import"

    if validate_project and context.project is None:
        raise NeptuneProjectNotProvided()
    if context.api_token is None:
        raise NeptuneApiTokenNotProvided()

    return context


def _context_from_env() -> Context:
    try:
        project = NEPTUNE_PROJECT.get()
    except ValueError:
        project = None

    try:
        api_token = NEPTUNE_API_TOKEN.get()
    except ValueError:
        api_token = None

    return Context(project=project, api_token=api_token)


_context: Context = _context_from_env()
_lock = threading.RLock()



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from typing import (
    Callable,
    Generic,
    Optional,
    TypeVar,
)

__all__ = (
    "NEPTUNE_API_TOKEN",
    "NEPTUNE_ENABLE_COLORS",
    "NEPTUNE_HTTP_REQUEST_TIMEOUT_SECONDS",
    "NEPTUNE_PROJECT",
    "NEPTUNE_QUERY_ATTRIBUTE_DEFINITIONS_BATCH_SIZE",
    "NEPTUNE_QUERY_ATTRIBUTE_VALUES_BATCH_SIZE",
    "NEPTUNE_QUERY_FILES_MAX_CONCURRENCY",
    "NEPTUNE_QUERY_FILES_TIMEOUT",
    "NEPTUNE_QUERY_MAX_REQUEST_SIZE",
    "NEPTUNE_QUERY_MAX_WORKERS",
    "NEPTUNE_QUERY_RETRY_HARD_TIMEOUT",
    "NEPTUNE_QUERY_RETRY_SOFT_TIMEOUT",
    "NEPTUNE_QUERY_SERIES_BATCH_SIZE",
    "NEPTUNE_QUERY_SYS_ATTRS_BATCH_SIZE",
    "NEPTUNE_VERIFY_SSL",
)

T = TypeVar("T")


class EnvVariable(Generic[T]):
    _UNSET: T = object()  # type: ignore

    def __init__(self, name: str, value_mapper: Callable[[str], T], default_value: T = _UNSET):
        self.name = name
        self._value_mapper = value_mapper
        self._default_value = default_value

    def get(self) -> T:
        value = os.getenv(self.name)
        if value is None:
            if self._default_value is self._UNSET:
                raise ValueError(f"Environment variable {self.name} is not set")
            return self._default_value
        return self._map_value(value)

    def _map_value(self, value: str) -> T:
        return self._value_mapper(value)


def _map_str(value: str) -> str:
    return value.strip()


def _map_bool(value: str) -> bool:
    return value.lower() in {"true", "1"}


def _lift_optional(mapper: Callable[[str], T]) -> Callable[[str], Optional[T]]:
    def wrapped(value: str) -> Optional[T]:
        if not value:
            return None
        return mapper(value)

    return wrapped


def _map_logging_level(value: str) -> str:
    valid_levels = {"CRITICAL", "FATAL", "ERROR", "WARN", "WARNING", "INFO", "DEBUG", "NOTSET"}
    level = value.strip().upper()
    if level not in valid_levels:
        return "WARN"
    return level


NEPTUNE_API_TOKEN = EnvVariable[str]("NEPTUNE_API_TOKEN", _map_str)
NEPTUNE_ENABLE_COLORS = EnvVariable[bool]("NEPTUNE_ENABLE_COLORS", _map_bool, True)
NEPTUNE_HTTP_REQUEST_TIMEOUT_SECONDS = EnvVariable[int]("NEPTUNE_HTTP_REQUEST_TIMEOUT_SECONDS", int, 60)
NEPTUNE_PROJECT = EnvVariable[str]("NEPTUNE_PROJECT", _map_str)
NEPTUNE_QUERY_ATTRIBUTE_DEFINITIONS_BATCH_SIZE = EnvVariable[int](
    "NEPTUNE_QUERY_ATTRIBUTE_DEFINITIONS_BATCH_SIZE", int, 10_000
)
NEPTUNE_QUERY_ATTRIBUTE_VALUES_BATCH_SIZE = EnvVariable[int]("NEPTUNE_QUERY_ATTRIBUTE_VALUES_BATCH_SIZE", int, 10_000)
NEPTUNE_QUERY_FILES_BATCH_SIZE = EnvVariable[int]("NEPTUNE_QUERY_FILES_BATCH_SIZE", int, 200)
NEPTUNE_QUERY_FILES_MAX_CONCURRENCY = EnvVariable[int]("NEPTUNE_QUERY_FILES_MAX_CONCURRENCY", int, 1)
NEPTUNE_QUERY_FILES_TIMEOUT = EnvVariable[Optional[int]]("NEPTUNE_QUERY_FILES_TIMEOUT", _lift_optional(int), None)
NEPTUNE_QUERY_MAX_REQUEST_SIZE = EnvVariable[int]("NEPTUNE_QUERY_MAX_REQUEST_SIZE", int, 220_000)
NEPTUNE_QUERY_MAX_WORKERS = EnvVariable[int]("NEPTUNE_QUERY_MAX_WORKERS", int, 32)
NEPTUNE_QUERY_RETRY_HARD_TIMEOUT = EnvVariable[Optional[int]](
    "NEPTUNE_QUERY_RETRY_HARD_TIMEOUT", _lift_optional(int), 3600
)
NEPTUNE_QUERY_RETRY_SOFT_TIMEOUT = EnvVariable[Optional[int]](
    "NEPTUNE_QUERY_RETRY_SOFT_TIMEOUT", _lift_optional(int), 1800
)
NEPTUNE_QUERY_SERIES_BATCH_SIZE = EnvVariable[int]("NEPTUNE_QUERY_SERIES_BATCH_SIZE", int, 10_000)
NEPTUNE_QUERY_SYS_ATTRS_BATCH_SIZE = EnvVariable[int]("NEPTUNE_QUERY_EXPERIMENT_SYS_ATTRS_BATCH_SIZE", int, 10_000)
NEPTUNE_VERIFY_SSL = EnvVariable[bool]("NEPTUNE_VERIFY_SSL", _map_bool, True)
NEPTUNE_LOGGER_LEVEL = EnvVariable[str]("NEPTUNE_LOGGER_LEVEL", _map_logging_level, "WARN")



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import os
import pathlib
from typing import (
    Iterable,
    Literal,
    Optional,
    Tuple,
)

from .. import filters
from ..retrieval.attribute_types import ATTRIBUTE_LITERAL


def restrict_attribute_filter_type(
    attribute_filter: filters._BaseAttributeFilter,
    type_in: Iterable[ATTRIBUTE_LITERAL],
) -> filters._BaseAttributeFilter:
    def restrict_type(leaf: filters._AttributeFilter) -> filters._AttributeFilter:
        # user: [A], type_in: [A, B] => [A]  # it's ok for user to request less types
        # user: [A, B], type_in: [A] => [A]  # remove types that are not supported
        # user: [A], type_in: [B] => raise   # raise an error when there is a complete mismatch
        intersection = [t for t in leaf.type_in if t in type_in]
        if intersection:
            return dataclasses.replace(leaf, type_in=intersection)
        else:
            types = list(type_in)
            if len(types) == 1:
                label = f"{types[0]} type is"
            else:
                label = f"{', '.join(types[:-1])} or {types[-1]} types are"

            raise ValueError(
                f"Only {label} supported for attribute filters in this function "
                f"and the filter contains types: {', '.join(leaf.type_in)}"
            )

    return attribute_filter.transform(map_attribute_filter=restrict_type)


def validate_include_time(include_time: Optional[Literal["absolute"]]) -> None:
    if include_time is not None and include_time not in ["absolute"]:
        raise ValueError("include_time must be 'absolute'")


def validate_step_range(step_range: Tuple[Optional[float], Optional[float]]) -> None:
    """Validate that a step range tuple contains valid values and is properly ordered."""
    if not isinstance(step_range, tuple) or len(step_range) != 2:
        raise ValueError("step_range must be a tuple of two values")

    start, end = step_range

    # Validate types
    if start is not None and not isinstance(start, (int, float)):
        raise ValueError("step_range start must be None or a number")
    if end is not None and not isinstance(end, (int, float)):
        raise ValueError("step_range end must be None or a number")

    # Validate range order if both values are provided
    if start is not None and end is not None and start > end:
        raise ValueError("step_range start must be less than or equal to end")


def _validate_optional_positive_int(value: Optional[int], name: str) -> None:
    """Validate that value is either None or a positive integer, with a custom name for error messages."""
    if value is not None:
        if not isinstance(value, int):
            raise ValueError(f"{name} must be None or an integer")
        if value <= 0:
            raise ValueError(f"{name} must be greater than 0")


def validate_tail_limit(tail_limit: Optional[int]) -> None:
    """Validate that tail_limit is either None or a positive integer."""
    _validate_optional_positive_int(tail_limit, "tail_limit")


def validate_limit(limit: Optional[int]) -> None:
    """Validate that limit is either None or a positive integer."""
    _validate_optional_positive_int(limit, "limit")


def validate_bucket_limit(limit: int) -> None:
    if not isinstance(limit, int):
        raise ValueError("limit must be an integer")
    if limit <= 0:
        raise ValueError(f"limit must be a positive integer. Got: {limit}")
    if limit > 1000:
        raise ValueError(f"limit cannot be greater than 1000. Got: {limit}")


def validate_metrics_x(x: Literal["step"]) -> Literal["step"]:
    """Validate that x is 'step' (the only valid value for now)."""
    if x not in ("step",):
        raise ValueError(f"x '{x}' is invalid; must be 'step'")
    return x


def validate_sort_direction(sort_direction: Literal["asc", "desc"]) -> Literal["asc", "desc"]:
    """Validate that sort_direction is either 'asc' or 'desc'."""
    if sort_direction not in ("asc", "desc"):
        raise ValueError(f"sort_direction '{sort_direction}' is invalid; must be 'asc' or 'desc'")
    return sort_direction


def ensure_write_access(destination: pathlib.Path) -> None:
    if not destination.exists():
        destination.mkdir(parents=True, exist_ok=True)

    if not destination.is_dir():
        raise NotADirectoryError(f"Destination is not a directory: {destination}")

    if not os.access(destination, os.W_OK):
        raise PermissionError(f"No write access to the directory: {destination}")



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools as ft
from dataclasses import dataclass
from enum import Enum
from typing import (
    Any,
    Callable,
    Generator,
    List,
    Literal,
    Optional,
    Protocol,
    TypeVar,
)

from neptune_api.api.retrieval import search_leaderboard_entries_proto
from neptune_api.client import AuthenticatedClient
from neptune_api.models import SearchLeaderboardEntriesParamsDTO
from neptune_api.proto.neptune_pb.api.v1.model.leaderboard_entries_pb2 import ProtoLeaderboardEntriesSearchResultDTO

from neptune_query.internal.query_metadata_context import with_neptune_client_metadata

from .. import (
    env,
    identifiers,
)
from ..filters import (
    _Attribute,
    _Filter,
)
from ..logger import get_logger
from ..retrieval import (
    retry,
    util,
)
from ..retrieval.attribute_types import map_attribute_type_python_to_backend

logger = get_logger()

_DIRECTION_PYTHON_TO_BACKEND_MAP: dict[str, str] = {
    "asc": "ascending",
    "desc": "descending",
}


def _map_direction(direction: Literal["asc", "desc"]) -> str:
    return _DIRECTION_PYTHON_TO_BACKEND_MAP[direction]


T = TypeVar("T")


class ContainerType(Enum):
    RUN = "run"
    EXPERIMENT = "experiment"


class SysIdLabel(Protocol):
    @property
    def sys_id(self) -> identifiers.SysId:
        ...

    @property
    def label(self) -> str:
        ...


@dataclass(frozen=True)
class ExperimentSysAttrs:
    sys_id: identifiers.SysId
    sys_name: identifiers.SysName

    @property
    def label(self) -> str:
        return self.sys_name

    @staticmethod
    def attribute_names() -> List[str]:
        return ["sys/name", "sys/id"]

    @staticmethod
    def from_dict(data: dict[str, Any]) -> "ExperimentSysAttrs":
        return ExperimentSysAttrs(
            sys_name=identifiers.SysName(data["sys/name"]),
            sys_id=identifiers.SysId(data["sys/id"]),
        )


@dataclass(frozen=True)
class RunSysAttrs:
    sys_id: identifiers.SysId
    sys_custom_run_id: identifiers.CustomRunId

    @property
    def label(self) -> str:
        return self.sys_custom_run_id

    @staticmethod
    def attribute_names() -> List[str]:
        return ["sys/custom_run_id", "sys/id"]

    @staticmethod
    def from_dict(data: dict[str, Any]) -> "RunSysAttrs":
        return RunSysAttrs(
            sys_custom_run_id=identifiers.CustomRunId(data["sys/custom_run_id"]),
            sys_id=identifiers.SysId(data["sys/id"]),
        )


def _sys_id_from_dict(data: dict[str, Any]) -> identifiers.SysId:
    return identifiers.SysId(data["sys/id"])


class FetchSysAttrs(Protocol[T]):
    def __call__(
        self,
        client: AuthenticatedClient,
        project_identifier: identifiers.ProjectIdentifier,
        filter_: Optional[_Filter] = None,
        sort_by: _Attribute = _Attribute("sys/creation_time", type="datetime"),
        sort_direction: Literal["asc", "desc"] = "desc",
        limit: Optional[int] = None,
        batch_size: int = env.NEPTUNE_QUERY_SYS_ATTRS_BATCH_SIZE.get(),
        container_type: ContainerType = ContainerType.EXPERIMENT,
    ) -> Generator[util.Page[T], None, None]:
        ...


def _create_fetch_sys_attrs(
    attribute_names: List[str],
    make_record: Callable[[dict[str, Any]], T],
    default_container_type: ContainerType,
) -> FetchSysAttrs[T]:
    def fetch_sys_attrs(
        client: AuthenticatedClient,
        project_identifier: identifiers.ProjectIdentifier,
        filter_: Optional[_Filter] = None,
        sort_by: _Attribute = _Attribute("sys/creation_time", type="datetime"),
        sort_direction: Literal["asc", "desc"] = "desc",
        limit: Optional[int] = None,
        batch_size: int = env.NEPTUNE_QUERY_SYS_ATTRS_BATCH_SIZE.get(),
        container_type: ContainerType = default_container_type,
    ) -> Generator[util.Page[T], None, None]:
        params: dict[str, Any] = {
            "attributeFilters": [{"path": attribute_name} for attribute_name in attribute_names],
            "pagination": {"limit": batch_size},
            "experimentLeader": container_type == ContainerType.EXPERIMENT,
            "sorting": {
                "dir": _map_direction(sort_direction),
                "sortBy": {"name": sort_by.name},
            },
        }
        if filter_ is not None:
            params["query"] = {"query": str(filter_)}
        if sort_by.aggregation is not None:
            params["sorting"]["aggregationMode"] = sort_by.aggregation
        if sort_by.type is not None:
            params["sorting"]["sortBy"]["type"] = map_attribute_type_python_to_backend(sort_by.type)

        return util.fetch_pages(
            client=client,
            fetch_page=ft.partial(_fetch_sys_attrs_page, project_identifier=project_identifier),
            process_page=ft.partial(_process_sys_attrs_page, make_record=make_record),
            make_new_page_params=ft.partial(_make_new_sys_attrs_page_params, batch_size=batch_size, limit=limit),
            params=params,
        )

    return fetch_sys_attrs


fetch_experiment_sys_attrs = _create_fetch_sys_attrs(
    attribute_names=ExperimentSysAttrs.attribute_names(),
    make_record=ExperimentSysAttrs.from_dict,
    default_container_type=ContainerType.EXPERIMENT,
)


fetch_run_sys_attrs = _create_fetch_sys_attrs(
    attribute_names=RunSysAttrs.attribute_names(),
    make_record=RunSysAttrs.from_dict,
    default_container_type=ContainerType.RUN,
)


def fetch_sys_id_labels(container_type: ContainerType) -> FetchSysAttrs[SysIdLabel]:
    if container_type == ContainerType.EXPERIMENT:
        return fetch_experiment_sys_attrs  # type: ignore
    elif container_type == ContainerType.RUN:
        return fetch_run_sys_attrs  # type: ignore
    else:
        raise RuntimeError(f"Unexpected container type: {container_type}")


fetch_experiment_sys_ids = _create_fetch_sys_attrs(
    attribute_names=["sys/id"], make_record=_sys_id_from_dict, default_container_type=ContainerType.EXPERIMENT
)

fetch_run_sys_ids = _create_fetch_sys_attrs(
    attribute_names=["sys/id"],
    make_record=_sys_id_from_dict,
    default_container_type=ContainerType.RUN,
)

fetch_sys_ids = fetch_experiment_sys_ids


def _fetch_sys_attrs_page(
    client: AuthenticatedClient,
    params: dict[str, Any],
    project_identifier: identifiers.ProjectIdentifier,
) -> ProtoLeaderboardEntriesSearchResultDTO:
    logger.debug(
        f"Calling search_leaderboard_entries_proto with project_identifier: {project_identifier}, params: {params}"
    )

    body = SearchLeaderboardEntriesParamsDTO.from_dict(params)
    call_api = retry.handle_errors_default(with_neptune_client_metadata(search_leaderboard_entries_proto.sync_detailed))
    response = call_api(client=client, project_identifier=project_identifier, type=["run"], body=body)

    logger.debug(
        f"search_leaderboard_entries_proto response status: {response.status_code}, "
        f"content length: {len(response.content) if response.content else 'no content'}"
    )
    return ProtoLeaderboardEntriesSearchResultDTO.FromString(response.content)


def _process_sys_attrs_page(
    data: ProtoLeaderboardEntriesSearchResultDTO,
    make_record: Callable[[dict[str, Any]], T],
) -> util.Page[T]:
    items = []
    for entry in data.entries:
        attributes = {attr.name: attr.string_properties.value for attr in entry.attributes}
        item = make_record(attributes)
        items.append(item)
    return util.Page(items=items)


def _make_new_sys_attrs_page_params(
    params: dict[str, Any],
    data: Optional[ProtoLeaderboardEntriesSearchResultDTO],
    batch_size: int,
    limit: Optional[int],
) -> Optional[dict[str, Any]]:

    if data is None:
        params["pagination"]["offset"] = 0
        if limit is not None:
            params["pagination"]["limit"] = min(limit, batch_size)
        return params

    if len(data.entries) < batch_size:
        return None

    params["pagination"]["offset"] += batch_size
    if limit is not None:
        offset = params["pagination"]["offset"]
        params["pagination"]["limit"] = min(limit - offset, batch_size)

    return params



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools as ft
from dataclasses import dataclass
from typing import (
    Any,
    Generator,
    Iterable,
    Optional,
    Union,
)

from neptune_api.api.retrieval import query_attributes_within_project_proto
from neptune_api.client import AuthenticatedClient
from neptune_api.models import QueryAttributesBodyDTO
from neptune_api.proto.neptune_pb.api.v1.model.attributes_pb2 import ProtoQueryAttributesResultDTO

from neptune_query.internal.query_metadata_context import with_neptune_client_metadata

from .. import (
    env,
    filters,
    identifiers,
)
from ..logger import get_logger
from ..retrieval import (
    retry,
    util,
)
from ..retrieval.attribute_types import (
    extract_value,
    map_attribute_type_backend_to_python,
)
from .attribute_filter import transform_attribute_filter_into_params

logger = get_logger()


@dataclass(frozen=True)
class AttributeValue:
    attribute_definition: identifiers.AttributeDefinition
    value: Any
    run_identifier: identifiers.RunIdentifier


def fetch_attribute_values(
    client: AuthenticatedClient,
    project_identifier: identifiers.ProjectIdentifier,
    run_identifiers: Iterable[identifiers.RunIdentifier],
    attribute_definitions: Union[Iterable[identifiers.AttributeDefinition], filters._AttributeFilter],
    batch_size: int = env.NEPTUNE_QUERY_ATTRIBUTE_VALUES_BATCH_SIZE.get(),
) -> Generator[util.Page[AttributeValue], None, None]:
    attribute_definitions_set: Optional[set[identifiers.AttributeDefinition]] = None

    params: dict[str, Any] = {
        "nextPage": {"limit": batch_size},
    }
    if run_identifiers is not None:
        params["experimentIdsFilter"] = [str(e) for e in run_identifiers]

    if isinstance(attribute_definitions, filters._AttributeFilter):
        attribute_filter_params = transform_attribute_filter_into_params(attribute_definitions)
        params.update(attribute_filter_params)
    else:
        attribute_definitions_set = set(attribute_definitions)
        if not attribute_definitions_set or not run_identifiers:
            yield from []
            return
        params["attributeNamesFilter"] = [ad.name for ad in attribute_definitions_set]

    yield from util.fetch_pages(
        client=client,
        fetch_page=ft.partial(_fetch_attribute_values_page, project_identifier=project_identifier),
        process_page=ft.partial(
            _process_attribute_values_page,
            attribute_definitions_set=attribute_definitions_set,
            project_identifier=project_identifier,
        ),
        make_new_page_params=_make_new_attribute_values_page_params,
        params=params,
    )


def _fetch_attribute_values_page(
    client: AuthenticatedClient,
    params: dict[str, Any],
    project_identifier: identifiers.ProjectIdentifier,
) -> ProtoQueryAttributesResultDTO:
    logger.debug(
        f"Calling query_attributes_within_project_proto with project_identifier: {project_identifier}, params: {params}"
    )

    body = QueryAttributesBodyDTO.from_dict(params)
    call_api = retry.handle_errors_default(
        with_neptune_client_metadata(query_attributes_within_project_proto.sync_detailed)
    )
    response = call_api(client=client, body=body, project_identifier=project_identifier)

    logger.debug(
        f"query_attributes_within_project_proto response status: {response.status_code}, "
        f"content length: {len(response.content) if response.content else 'no content'}"
    )
    return ProtoQueryAttributesResultDTO.FromString(response.content)


def _process_attribute_values_page(
    data: ProtoQueryAttributesResultDTO,
    attribute_definitions_set: Optional[set[identifiers.AttributeDefinition]],
    project_identifier: identifiers.ProjectIdentifier,
) -> util.Page[AttributeValue]:
    items = []
    for entry in data.entries:
        run_identifier = identifiers.RunIdentifier(
            project_identifier=project_identifier, sys_id=identifiers.SysId(entry.experimentShortId)
        )

        for attr in entry.attributes:
            attr_definition = identifiers.AttributeDefinition(
                name=attr.name, type=map_attribute_type_backend_to_python(attr.type)
            )
            if attribute_definitions_set is not None and attr_definition not in attribute_definitions_set:
                continue

            item_value = extract_value(attr)
            if item_value is None:
                continue

            attr_value = AttributeValue(
                attribute_definition=attr_definition,
                value=item_value,
                run_identifier=run_identifier,
            )
            items.append(attr_value)

    return util.Page(items=items)


def _make_new_attribute_values_page_params(
    params: dict[str, Any], data: Optional[ProtoQueryAttributesResultDTO]
) -> Optional[dict[str, Any]]:
    if data is None:
        if "nextPageToken" in params["nextPage"]:
            del params["nextPage"]["nextPageToken"]
        return params

    next_page_token = data.nextPage.nextPageToken
    if not next_page_token:
        return None

    params["nextPage"]["nextPageToken"] = next_page_token
    return params



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import (
    Generator,
    Literal,
    Optional,
    Tuple,
)

import pandas as pd

from .. import identifiers
from ..client import get_client
from ..composition import attribute_components as _components
from ..composition import (
    concurrency,
    type_inference,
    validation,
)
from ..context import (
    Context,
    get_context,
    validate_context,
)
from ..filters import (
    _BaseAttributeFilter,
    _Filter,
)
from ..identifiers import ProjectIdentifier
from ..output_format import create_series_dataframe
from ..retrieval import (
    search,
    series,
    split,
    util,
)
from ..retrieval.search import ContainerType

__all__ = ("fetch_series",)


def fetch_series(
    *,
    project_identifier: ProjectIdentifier,
    filter_: Optional[_Filter],
    attributes: _BaseAttributeFilter,
    include_time: Optional[Literal["absolute"]],
    step_range: Tuple[Optional[float], Optional[float]],
    lineage_to_the_root: bool,
    tail_limit: Optional[int],
    context: Optional[Context] = None,
    container_type: ContainerType,
) -> pd.DataFrame:
    validation.validate_step_range(step_range)
    validation.validate_tail_limit(tail_limit)
    validation.validate_include_time(include_time)
    attributes_restricted = validation.restrict_attribute_filter_type(
        attributes, type_in={"string_series", "histogram_series", "file_series"}
    )

    valid_context = validate_context(context or get_context())
    client = get_client(context=valid_context)

    with (
        concurrency.create_thread_pool_executor() as executor,
        concurrency.create_thread_pool_executor() as fetch_attribute_definitions_executor,
    ):
        inference_result = type_inference.infer_attribute_types_in_filter(
            client=client,
            project_identifier=project_identifier,
            filter_=filter_,
            fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
        )
        inferred_filter = inference_result.get_result_or_raise()
        inference_result.emit_warnings()

        sys_id_label_mapping: dict[identifiers.SysId, str] = {}

        def go_fetch_sys_attrs() -> Generator[list[identifiers.SysId], None, None]:
            for page in search.fetch_sys_id_labels(container_type)(
                client=client,
                project_identifier=project_identifier,
                filter_=inferred_filter,
            ):
                sys_ids = []
                for item in page.items:
                    sys_id_label_mapping[item.sys_id] = item.label
                    sys_ids.append(item.sys_id)
                yield sys_ids

        output = concurrency.generate_concurrently(
            items=go_fetch_sys_attrs(),
            executor=executor,
            downstream=lambda sys_ids: _components.fetch_attribute_definitions_split(
                client=client,
                project_identifier=project_identifier,
                attribute_filter=attributes_restricted,
                executor=executor,
                fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
                sys_ids=sys_ids,
                downstream=lambda sys_ids_split, definitions_page: concurrency.generate_concurrently(
                    items=split.split_series_attributes(
                        items=(
                            identifiers.RunAttributeDefinition(
                                run_identifier=identifiers.RunIdentifier(project_identifier, sys_id),
                                attribute_definition=definition,
                            )
                            for sys_id in sys_ids_split
                            for definition in definitions_page.items
                        ),
                    ),
                    executor=executor,
                    downstream=lambda run_attribute_definitions_split: concurrency.generate_concurrently(
                        items=series.fetch_series_values(
                            client=client,
                            run_attribute_definitions=run_attribute_definitions_split,
                            include_inherited=lineage_to_the_root,
                            container_type=container_type,
                            step_range=step_range,
                            tail_limit=tail_limit,
                        ),
                        executor=executor,
                        downstream=concurrency.return_value,
                    ),
                ),
            ),
        )
        results: Generator[
            util.Page[tuple[identifiers.RunAttributeDefinition, list[series.SeriesValue]]], None, None
        ] = concurrency.gather_results(output)

        series_data: dict[identifiers.RunAttributeDefinition, list[series.SeriesValue]] = {}
        for result in results:
            for run_attribute_definition, series_values in result.items:
                series_data.setdefault(run_attribute_definition, []).extend(series_values)

        return create_series_dataframe(
            series_data,
            project_identifier,
            sys_id_label_mapping,
            index_column_name="experiment" if container_type == ContainerType.EXPERIMENT else "run",
            timestamp_column_name="absolute_time" if include_time == "absolute" else None,
        )



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from dataclasses import dataclass
from io import BytesIO
from typing import (
    Any,
    Callable,
    Generator,
    Generic,
    Optional,
    TypeVar,
)

from google.protobuf.message import Message
from neptune_api import AuthenticatedClient
from neptune_api.types import File

T = TypeVar("T")
R = TypeVar("R")
_Params = dict[str, Any]


@dataclass
class Page(Generic[T]):
    items: list[T]


def fetch_pages(
    client: AuthenticatedClient,
    fetch_page: Callable[[AuthenticatedClient, _Params], R],
    process_page: Callable[[R], Page[T]],
    make_new_page_params: Callable[[_Params, Optional[R]], Optional[_Params]],
    params: _Params,
) -> Generator[Page[T], None, None]:
    page_params = make_new_page_params(params, None)
    while page_params is not None:
        data = fetch_page(client, page_params)
        page = process_page(data)
        yield page
        page_params = make_new_page_params(page_params, data)


class ProtobufPayload(File):
    """A version of the neptune_api.types.File class that uses a protobuf message as payload"""

    @property
    def payload(self) -> BytesIO:
        return self.get_payload()

    @payload.setter
    def payload(self, message: Message) -> None:
        self.get_payload = lambda: BytesIO(message.SerializeToString())



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from concurrent.futures import Executor
from typing import (
    Generator,
    Literal,
    Optional,
    Protocol,
    Union,
)

import numpy as np
import pandas as pd
from neptune_api.client import AuthenticatedClient

from .. import identifiers
from ..client import get_client
from ..composition import (
    concurrency,
    type_inference,
    validation,
)
from ..context import (
    Context,
    get_context,
    validate_context,
)
from ..filters import (
    _BaseAttributeFilter,
    _Filter,
)
from ..identifiers import RunAttributeDefinition
from ..output_format import create_metric_buckets_dataframe
from ..retrieval import (
    metric_buckets,
    search,
    util,
)
from ..retrieval.attribute_values import AttributeValue
from ..retrieval.metric_buckets import (
    MAX_SERIES_PER_REQUEST,
    TimeseriesBucket,
)
from ..retrieval.search import ContainerType
from .attribute_components import fetch_attribute_values_by_filter_split

__all__ = ("fetch_metric_buckets",)


class _FetchInChunksProtocol(Protocol):
    def __call__(
        self, x_range: Optional[tuple[float, float]], bucket_limit: int
    ) -> dict[RunAttributeDefinition, list[TimeseriesBucket]]:
        ...


def fetch_metric_buckets(
    *,
    project_identifier: identifiers.ProjectIdentifier,
    filter_: Optional[_Filter],
    x: Union[Literal["step"]] = "step",
    y: _BaseAttributeFilter,
    limit: int,
    lineage_to_the_root: bool,
    include_point_previews: bool,
    context: Optional[Context] = None,
    container_type: ContainerType,
) -> pd.DataFrame:
    validation.validate_metrics_x(x)
    validation.validate_bucket_limit(limit)
    restricted_y = validation.restrict_attribute_filter_type(y, type_in={"float_series"})
    limit = limit + 1  # we request one extra bucket because the 1st one is (-inf, 1st point] and we merge it

    valid_context = validate_context(context or get_context())
    client = get_client(context=valid_context)

    with (
        concurrency.create_thread_pool_executor() as executor,
        concurrency.create_thread_pool_executor() as fetch_attribute_definitions_executor,
    ):
        inference_result = type_inference.infer_attribute_types_in_filter(
            client=client,
            project_identifier=project_identifier,
            filter_=filter_,
            fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
        )
        inferred_filter = inference_result.get_result_or_raise()
        inference_result.emit_warnings()

        metric_buckets_data, sys_id_to_label_mapping = _fetch_metric_buckets(
            filter_=inferred_filter,
            x=x,
            y=restricted_y,
            client=client,
            project_identifier=project_identifier,
            limit=limit,
            lineage_to_the_root=lineage_to_the_root,
            include_point_previews=include_point_previews,
            executor=executor,
            fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
            container_type=container_type,
        )

        df = create_metric_buckets_dataframe(
            buckets_data=metric_buckets_data,
            sys_id_label_mapping=sys_id_to_label_mapping,
            container_column_name="experiment" if container_type == ContainerType.EXPERIMENT else "run",
        )

    return df


def _fetch_metric_buckets(
    filter_: Optional[_Filter],
    x: Literal["step"],
    y: _BaseAttributeFilter,
    client: AuthenticatedClient,
    project_identifier: identifiers.ProjectIdentifier,
    executor: Executor,
    fetch_attribute_definitions_executor: Executor,
    lineage_to_the_root: bool,
    include_point_previews: bool,
    limit: int,
    container_type: ContainerType,
) -> tuple[dict[identifiers.RunAttributeDefinition, list[TimeseriesBucket]], dict[identifiers.SysId, str]]:
    sys_id_label_mapping: dict[identifiers.SysId, str] = {}

    def go_fetch_sys_attrs() -> Generator[list[identifiers.SysId], None, None]:
        for page in search.fetch_sys_id_labels(container_type)(
            client=client,
            project_identifier=project_identifier,
            filter_=filter_,
        ):
            sys_ids = []
            for item in page.items:
                sys_id_label_mapping[item.sys_id] = item.label
                sys_ids.append(item.sys_id)
            yield sys_ids

    output = concurrency.generate_concurrently(
        items=go_fetch_sys_attrs(),
        executor=executor,
        downstream=lambda sys_ids: fetch_attribute_values_by_filter_split(
            client=client,
            project_identifier=project_identifier,
            sys_ids=sys_ids,
            attribute_filter=y,
            executor=executor,
            fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
            downstream=concurrency.return_value,
        ),
    )

    attribute_value_pages: Generator[util.Page[AttributeValue], None, None] = concurrency.gather_results(output)

    run_attribute_definitions = []
    for page in attribute_value_pages:
        for value in page.items:
            run_attribute_definition = RunAttributeDefinition(
                run_identifier=value.run_identifier, attribute_definition=value.attribute_definition
            )
            run_attribute_definitions.append(run_attribute_definition)

    if not run_attribute_definitions:
        return {}, sys_id_label_mapping

    fetch_in_chunks: _FetchInChunksProtocol = lambda x_range, bucket_limit: _fetch_in_chunks(
        client=client,
        run_attribute_definitions=run_attribute_definitions,
        x=x,
        include_point_previews=include_point_previews,
        container_type=container_type,
        x_range=x_range,
        lineage_to_the_root=lineage_to_the_root,
        limit=bucket_limit,
        executor=executor,
    )

    # if len(run_attribute_definitions) <= MAX_SERIES_PER_REQUEST:
    #     return fetch_in_chunks(x_range=None, bucket_limit=limit), sys_id_label_mapping

    global_x_range = _compute_global_x_range(fetch_in_chunks=fetch_in_chunks)
    if global_x_range is None:
        # No finite points / bucket bounds found
        return {}, sys_id_label_mapping

    return fetch_in_chunks(x_range=global_x_range, bucket_limit=limit), sys_id_label_mapping


def _fetch_in_chunks(
    client: AuthenticatedClient,
    run_attribute_definitions: list[RunAttributeDefinition],
    x: Literal["step"],
    include_point_previews: bool,
    container_type: ContainerType,
    x_range: Optional[tuple[float, float]],
    lineage_to_the_root: bool,
    limit: int,
    executor: Executor,
) -> dict[RunAttributeDefinition, list[TimeseriesBucket]]:
    chunks = (
        run_attribute_definitions[offset : offset + MAX_SERIES_PER_REQUEST]
        for offset in range(0, len(run_attribute_definitions), MAX_SERIES_PER_REQUEST)
    )

    output = concurrency.generate_concurrently(
        items=chunks,
        executor=executor,
        downstream=lambda chunk: concurrency.return_value(
            metric_buckets.fetch_time_series_buckets(
                client=client,
                x=x,
                run_attribute_definitions=chunk,
                lineage_to_the_root=lineage_to_the_root,
                include_point_previews=include_point_previews,
                limit=limit,
                container_type=container_type,
                x_range=x_range,
            )
        ),
    )

    merged: dict[RunAttributeDefinition, list[TimeseriesBucket]] = {}
    for chunk_data in concurrency.gather_results(output):  # type: dict[RunAttributeDefinition, list[TimeseriesBucket]]
        merged.update(chunk_data)
    return merged


def _compute_global_x_range(fetch_in_chunks: _FetchInChunksProtocol) -> Optional[tuple[float, float]]:
    x_range: tuple[Optional[float], Optional[float]] = (None, None)
    # We only need the minimal number of buckets to determine min/max x
    for buckets in fetch_in_chunks(x_range=None, bucket_limit=2).values():
        for bucket in buckets:
            x_range = _update_range(x_range, bucket)

    if x_range[0] is None or x_range[1] is None:
        return None
    return x_range[0], x_range[1]


def _update_range(
    current_range: tuple[Optional[float], Optional[float]], bucket: TimeseriesBucket
) -> tuple[Optional[float], Optional[float],]:
    # We're including from_x and to_x because some buckets might hold only non-finite points,
    # in which case first_x and last_x are None.
    candidates = [bucket.first_x, bucket.last_x, bucket.from_x, bucket.to_x] + list(current_range)
    finite_candidates = [x for x in candidates if x is not None and np.isfinite(x)]

    if len(finite_candidates):
        return min(finite_candidates), max(finite_candidates)
    else:
        return None, None



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import (
    Generator,
    Literal,
    Optional,
)

import pandas as pd

from .. import client as _client
from .. import context as _context
from .. import (
    identifiers,
    output_format,
)
from ..composition import attribute_components as _components
from ..composition import (
    concurrency,
    type_inference,
    validation,
)
from ..filters import (
    _Attribute,
    _BaseAttributeFilter,
    _Filter,
)
from ..identifiers import ProjectIdentifier
from ..retrieval import attribute_values as att_vals
from ..retrieval import (
    search,
    util,
)

__all__ = ("fetch_table",)


def fetch_table(
    *,
    project_identifier: ProjectIdentifier,
    filter_: Optional[_Filter],
    attributes: _BaseAttributeFilter,
    sort_by: _Attribute,
    sort_direction: Literal["asc", "desc"],
    limit: Optional[int],
    type_suffix_in_column_names: bool,
    context: Optional[_context.Context] = None,
    container_type: search.ContainerType,
) -> pd.DataFrame:
    validation.validate_limit(limit)
    _sort_direction = validation.validate_sort_direction(sort_direction)

    valid_context = _context.validate_context(context or _context.get_context())
    client = _client.get_client(context=valid_context)

    with (
        concurrency.create_thread_pool_executor() as executor,
        concurrency.create_thread_pool_executor() as fetch_attribute_definitions_executor,
    ):

        inference_result = type_inference.infer_attribute_types_in_filter(
            client=client,
            project_identifier=project_identifier,
            filter_=filter_,
            fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
        )
        filter_ = inference_result.get_result_or_raise()
        inference_result.emit_warnings()

        sort_by_inference_result = type_inference.infer_attribute_types_in_sort_by(
            client=client,
            project_identifier=project_identifier,
            sort_by=sort_by,
            fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
        )
        sort_by = sort_by_inference_result.result
        sort_by_inference_result.emit_warnings()

        sys_id_label_mapping: dict[identifiers.SysId, str] = {}
        result_by_id: dict[identifiers.SysId, list[att_vals.AttributeValue]] = {}

        def go_fetch_sys_attrs() -> Generator[list[identifiers.SysId], None, None]:
            for page in search.fetch_sys_id_labels(container_type)(
                client=client,
                project_identifier=project_identifier,
                filter_=filter_,
                sort_by=sort_by,
                sort_direction=_sort_direction,
                limit=limit,
            ):
                sys_ids = []
                for item in page.items:
                    result_by_id[item.sys_id] = []  # I assume that dict preserves the order set here
                    sys_id_label_mapping[item.sys_id] = item.label
                    sys_ids.append(item.sys_id)
                yield sys_ids

        output = concurrency.generate_concurrently(
            items=go_fetch_sys_attrs(),
            executor=executor,
            downstream=lambda sys_ids: _components.fetch_attribute_definitions_split(
                client=client,
                project_identifier=project_identifier,
                attribute_filter=attributes,
                executor=executor,
                fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
                sys_ids=sys_ids,
                downstream=lambda sys_ids_split, definitions_page: _components.fetch_attribute_values_split(
                    client=client,
                    project_identifier=project_identifier,
                    executor=executor,
                    sys_ids=sys_ids_split,
                    attribute_definitions=definitions_page.items,
                    downstream=concurrency.return_value,
                ),
            ),
        )
        results: Generator[util.Page[att_vals.AttributeValue], None, None] = concurrency.gather_results(output)

        for result in results:
            attribute_values_page = result
            for attribute_value in attribute_values_page.items:
                sys_id = attribute_value.run_identifier.sys_id
                result_by_id[sys_id].append(attribute_value)

    result_by_name = _map_keys_preserving_order(result_by_id, sys_id_label_mapping)
    dataframe = output_format.convert_table_to_dataframe(
        table_data=result_by_name,
        project_identifier=project_identifier,
        type_suffix_in_column_names=type_suffix_in_column_names,
        index_column_name="experiment" if container_type == search.ContainerType.EXPERIMENT else "run",
    )

    return dataframe


def _map_keys_preserving_order(
    result_by_id: dict[identifiers.SysId, list[att_vals.AttributeValue]],
    sys_id_label_mapping: dict[identifiers.SysId, str],
) -> dict[str, list[att_vals.AttributeValue]]:
    result_by_name = {}
    for sys_id, values in result_by_id.items():
        label = sys_id_label_mapping[sys_id]
        result_by_name[label] = values
    return result_by_name



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

import contextlib
import dataclasses
import functools
import json
from dataclasses import dataclass
from importlib.metadata import (
    PackageNotFoundError,
    version,
)
from typing import (
    Callable,
    Generator,
    Optional,
    ParamSpec,
    TypeVar,
)

from neptune_api.types import Response

from neptune_query.internal.composition import concurrency

# This flag is used to control whether query metadata should be added to the request headers
# Can remove this after backend supports this properly (under ticket PY-171)
ADD_QUERY_METADATA = False


@functools.cache
def get_client_version() -> str:
    package_name = "neptune-query"
    try:
        package_version = version(package_name)
    except PackageNotFoundError:
        package_version = "unknown"
    return f"{package_name}/{package_version}"


@dataclass
class QueryMetadata:
    api_function: str
    client_version: str

    def __post_init__(self) -> None:
        self.api_function = self.api_function[:50]
        self.client_version = self.client_version[:50]


@contextlib.contextmanager
def use_query_metadata(api_function: str) -> Generator[None, None, None]:
    query_metadata = QueryMetadata(
        api_function=api_function,
        client_version=get_client_version(),
    )
    with concurrency.use_thread_local({"query_metadata": query_metadata}):
        yield


T = ParamSpec("T")
R = TypeVar("R")


def with_neptune_client_metadata(func: Callable[T, Response[R]]) -> Callable[T, Response[R]]:
    @functools.wraps(func)
    def wrapper(*args: T.args, **kwargs: T.kwargs) -> Response[R]:
        query_metadata: Optional[QueryMetadata] = concurrency.get_thread_local(
            "query_metadata", expected_type=QueryMetadata
        )
        if ADD_QUERY_METADATA and query_metadata:
            kwargs["x_neptune_client_metadata"] = json.dumps(dataclasses.asdict(query_metadata))
        return func(*args, **kwargs)

    return wrapper



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import (
    Iterable,
    Literal,
    Optional,
)

from neptune_api.api.retrieval import get_timeseries_buckets_proto
from neptune_api.client import AuthenticatedClient
from neptune_api.proto.neptune_pb.api.v1.model.requests_pb2 import (
    LineageEntityType,
    ProtoCustomExpression,
    ProtoGetTimeseriesBucketsRequest,
    ProtoLineage,
    ProtoScale,
    ProtoView,
    ProtoXAxis,
    XSteps,
)
from neptune_api.proto.neptune_pb.api.v1.model.series_values_pb2 import ProtoTimeseriesBucketsDTO

from ..identifiers import RunAttributeDefinition
from ..logger import get_logger
from ..query_metadata_context import with_neptune_client_metadata
from . import retry
from .search import ContainerType
from .util import ProtobufPayload

logger = get_logger()


MAX_SERIES_PER_REQUEST = 1000


@dataclass(frozen=True)
class TimeseriesBucket:
    index: int
    from_x: float
    to_x: float
    first_x: Optional[float]
    first_y: Optional[float]
    last_x: Optional[float]
    last_y: Optional[float]

    # statistics:
    # y_min: Optional[float]
    # y_max: Optional[float]
    # finite_point_count: int
    # nan_count: int
    # positive_inf_count: int
    # negative_inf_count: int
    # finite_points_sum: Optional[float]


# Build once at module import
ATTRIBUTE_NAME_TO_FORMULA_ESCAPE_MAP = {
    ord("\\"): r"\\",
    ord("$"): r"\$",
    ord("{"): r"\{",
    ord("}"): r"\}",
}


def attribute_name_to_formula(attribute_name: str) -> str:
    return "${" + attribute_name.translate(ATTRIBUTE_NAME_TO_FORMULA_ESCAPE_MAP) + "}"


def int_to_uuid(num: int) -> str:
    # Generate a UUID from an integer in a deterministic way
    # This is a simple implementation and can be replaced with a more robust one if needed
    return f"00000000-0000-0000-0000-{num:012d}"


def fetch_time_series_buckets(
    client: AuthenticatedClient,
    run_attribute_definitions: Iterable[RunAttributeDefinition],
    container_type: ContainerType,
    x: Literal["step"],
    lineage_to_the_root: bool,
    include_point_previews: bool,
    limit: int,
    x_range: Optional[tuple[float, float]],
) -> dict[RunAttributeDefinition, list[TimeseriesBucket]]:
    run_attribute_definitions = list(run_attribute_definitions)

    lineage = ProtoLineage.FULL if lineage_to_the_root else ProtoLineage.ONLY_OWNED
    lineage_entity_type = LineageEntityType.RUN if container_type == ContainerType.RUN else LineageEntityType.EXPERIMENT

    if x == "step":
        xAxis = ProtoXAxis(steps=XSteps())
    # elif x == "relativeTime":
    #     xAxis = ProtoXAxis(relativeTime=XRelativeTime())
    # elif x == "epochMillis":
    #     xAxis = ProtoXAxis(epochMillis=XEpochMillis())
    # elif x == "custom":
    #     xAxis = ProtoXAxis(custom=XCustom())
    else:
        raise ValueError('Unsupported x value. Only "step" is supported')

    expressions = {}
    request_id_to_request_mapping = {}
    for num, run_attribute_definition in enumerate(run_attribute_definitions):
        if num >= MAX_SERIES_PER_REQUEST:
            raise ValueError(f"Cannot fetch more than {MAX_SERIES_PER_REQUEST} time series at once")

        request_id = int_to_uuid(num)
        request_id_to_request_mapping[request_id] = run_attribute_definition

        expressions[request_id] = ProtoCustomExpression(
            requestId=request_id,
            runId=str(run_attribute_definition.run_identifier),
            entityType=lineage_entity_type,
            lineage=lineage,
            customYFormula=attribute_name_to_formula(run_attribute_definition.attribute_definition.name),
            includePreview=include_point_previews,
        )

    view = ProtoView(
        maxBuckets=limit,
        xScale=ProtoScale.linear,
        yScale=ProtoScale.linear,
        xAxis=xAxis,
    )
    if x_range is not None:
        x_from, x_to = x_range
        view.to = x_to
        setattr(view, "from", x_from)  # from is a reserved keyword in Python

    request_object = ProtoGetTimeseriesBucketsRequest(
        expressions=expressions.values(),
        view=view,
    )

    logger.debug(f"Calling get_timeseries_buckets_proto with body: {request_object}")

    call_api = retry.handle_errors_default(with_neptune_client_metadata(get_timeseries_buckets_proto.sync_detailed))
    response = call_api(
        client=client,
        body=ProtobufPayload(request_object),
    )

    logger.debug(
        f"get_timeseries_buckets_proto response status: {response.status_code}, "
        f"content length: {len(response.content) if response.content else 'no content'}"
    )

    result_object: ProtoTimeseriesBucketsDTO = ProtoTimeseriesBucketsDTO.FromString(response.content)

    out: dict[RunAttributeDefinition, list[TimeseriesBucket]] = {}

    for entry in result_object.entries:
        request = request_id_to_request_mapping.get(entry.requestId, None)
        if request is None:
            raise RuntimeError(f"Received unknown requestId from the server: {entry.requestId}")

        if request in out:
            raise RuntimeError(f"Received duplicate requestId from the server: {entry.requestId}")

        out[request] = [
            TimeseriesBucket(
                index=bucket.index,
                from_x=bucket.fromX,
                to_x=bucket.toX,
                first_x=bucket.first.x if bucket.HasField("first") else None,
                first_y=bucket.first.y if bucket.HasField("first") else None,
                last_x=bucket.last.x if bucket.HasField("last") else None,
                last_y=bucket.last.y if bucket.HasField("last") else None,
                # y_min=bucket.localMin if bucket.HasField("localMin") else None,
                # y_max=bucket.localMax if bucket.HasField("localMax") else None,
                # finite_point_count=bucket.finitePointCount,
                # nan_count=bucket.nanCount,
                # positive_inf_count=bucket.positiveInfCount,
                # negative_inf_count=bucket.negativeInfCount,
                # finite_points_sum=bucket.localSum if bucket.HasField("localSum") else None,
            )
            for bucket in entry.bucket
        ]

    for request in run_attribute_definitions:
        if request not in out:
            raise RuntimeError("Didn't get data for all the requests from the server. " f"Missing request {request}")

    return out



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import hashlib
import mimetypes
import pathlib
from dataclasses import dataclass
from typing import (
    Literal,
    Optional,
    Union,
)

import azure.core.exceptions
import requests
from azure.storage.blob import BlobClient as AzureBlobClient
from neptune_api.api.storage import signed_url_generic
from neptune_api.client import AuthenticatedClient
from neptune_api.models import (
    CreateSignedUrlsRequest,
    CreateSignedUrlsResponse,
    FileToSign,
    Permission,
    Provider,
)

from neptune_query.internal.query_metadata_context import with_neptune_client_metadata

from ...exceptions import NeptuneFileDownloadError
from ...types import File
from .. import (
    env,
    identifiers,
)
from ..logger import get_logger
from ..retrieval import retry

logger = get_logger()


@dataclass(frozen=True)
class SignedFile:
    url: str
    path: str
    provider: Literal["azure", "gcp", "aws"]
    project_identifier: identifiers.ProjectIdentifier
    permission: Literal["read", "write"]


def fetch_signed_urls(
    client: AuthenticatedClient,
    files: list[File] | list[SignedFile],
    permission: Literal["read", "write"] = "read",
) -> list[SignedFile]:
    body = CreateSignedUrlsRequest(
        files=[
            FileToSign(project_identifier=file.project_identifier, path=file.path, permission=Permission(permission))
            for file in files
        ]
    )

    logger.debug(f"Calling signed_url_generic with body: {body}")

    call_api = retry.handle_errors_default(with_neptune_client_metadata(signed_url_generic.sync_detailed))
    response = call_api(client=client, body=body)

    logger.debug(
        f"signed_url_generic response status: {response.status_code}, "
        f"content length: {len(response.content) if response.content else 'no content'}"
    )

    data: CreateSignedUrlsResponse = response.parsed
    if len(data.files) != len(files):
        missing_paths = {f.path for f in files} - {f.path for f in data.files}
        raise ValueError(
            f"Server returned {len(data.files)} / {len(files)} signed urls. " f"Missing paths: {missing_paths}"
        )
    return [
        SignedFile(
            url=file_.url,
            path=file_.path,
            provider=_verify_provider(file_.provider),
            project_identifier=identifiers.ProjectIdentifier(file_.project_identifier),
            permission=permission,
        )
        for file_ in data.files
    ]


def _verify_provider(provider: Provider) -> Literal["azure", "gcp", "aws"]:
    if provider == Provider.AZURE:
        return "azure"
    elif provider == Provider.GCP:
        return "gcp"
    elif provider == Provider.AWS:
        return "aws"
    else:
        raise ValueError(f"Unsupported storage provider: {provider}")


def refresh_signed_file(
    client: AuthenticatedClient,
    signed_file: SignedFile,
) -> SignedFile:
    """
    Refreshes the signed file URL by fetching a new signed URL from the server.
    This is useful when the original signed URL has expired or is no longer valid.
    """
    new_signed_files = fetch_signed_urls(client=client, files=[signed_file], permission=signed_file.permission)
    return new_signed_files[0]


@dataclass
class DownloadResult:
    status: Literal["success", "not_found", "expired", "transient"]
    status_code: Optional[int] = None
    content: Optional[Union[str, bytes]] = None


def download_file(
    signed_file: SignedFile,
    target_path: pathlib.Path,
    max_concurrency: int = env.NEPTUNE_QUERY_FILES_MAX_CONCURRENCY.get(),
    timeout: Optional[int] = env.NEPTUNE_QUERY_FILES_TIMEOUT.get(),
) -> DownloadResult:
    logger.info(f"Downloading file from {signed_file.provider} storage: {signed_file.url} to {target_path}")

    target_path.parent.mkdir(parents=True, exist_ok=True)

    if signed_file.provider == "azure":
        result = _download_file_azure(signed_file, target_path, max_concurrency, timeout)
    elif signed_file.provider == "gcp":
        result = _download_file_requests(signed_file, target_path, timeout)
    elif signed_file.provider == "aws":
        result = _download_file_requests(signed_file, target_path, timeout)
    else:
        raise ValueError(f"Unsupported provider: {signed_file.provider}")

    logger.debug(
        f"Download result for {signed_file.url}: status={result.status}, status_code={result.status_code}, "
        f"content={result.content!r}"
    )
    return result


def _download_file_azure(
    signed_file: SignedFile,
    target_path: pathlib.Path,
    max_concurrency: int = env.NEPTUNE_QUERY_FILES_MAX_CONCURRENCY.get(),
    timeout: Optional[int] = env.NEPTUNE_QUERY_FILES_TIMEOUT.get(),
) -> DownloadResult:
    try:
        blob_client = AzureBlobClient.from_blob_url(signed_file.url)
        download_stream = blob_client.download_blob(max_concurrency=max_concurrency, timeout=timeout)
        with open(target_path, mode="wb") as opened:
            for chunk in download_stream.chunks():
                opened.write(chunk)
        return DownloadResult(status="success")
    except azure.core.exceptions.ResourceNotFoundError as e:
        # This error indicates that the file does not exist in the storage - do not retry
        return DownloadResult(status="not_found", status_code=e.status_code, content=e.response.text())
    except azure.core.exceptions.ClientAuthenticationError as e:
        # This error indicates that the signed URL is no longer valid, likely due to expiration - try signing again
        return DownloadResult(status="expired", status_code=e.status_code, content=e.response.text())
    except azure.core.exceptions.HttpResponseError as e:
        # This error indicates a general HTTP error, which may be transient - retrying might help
        return DownloadResult(status="transient", status_code=e.status_code, content=e.response.text())


def _download_file_requests(
    signed_file: SignedFile,
    target_path: pathlib.Path,
    timeout: Optional[int] = env.NEPTUNE_QUERY_FILES_TIMEOUT.get(),
) -> DownloadResult:
    try:
        response = requests.get(signed_file.url, stream=True, timeout=timeout)
        response.raise_for_status()
        with open(target_path, mode="wb") as file:
            for chunk in response.iter_content(chunk_size=4 * 1024 * 1024):
                file.write(chunk)
        return DownloadResult(status="success")
    except requests.exceptions.HTTPError as e:
        try:
            if target_path.exists():
                target_path.unlink()
        except OSError:
            pass

        if e.response.status_code == 404:
            return DownloadResult(status="not_found", status_code=e.response.status_code, content=e.response.content)
        elif e.response.status_code == 400:
            return DownloadResult(status="expired", status_code=e.response.status_code, content=e.response.content)
        else:
            return DownloadResult(status="transient", status_code=e.response.status_code, content=e.response.content)


def download_file_complete(
    client: AuthenticatedClient,
    signed_file: SignedFile,
    target_path: pathlib.Path,
    max_tries: int = 3,
) -> Optional[pathlib.Path]:
    assert max_tries > 0, "max_tries must allow for at least one attempt"
    attempt = 1
    result = None
    try:
        while attempt <= max_tries:
            result = download_file(
                signed_file=signed_file,
                target_path=target_path,
            )
            if result.status == "success":
                return target_path
            elif result.status == "not_found":
                return None
            elif result.status == "expired":
                attempt += 1
                signed_file = refresh_signed_file(
                    client=client,
                    signed_file=signed_file,
                )
            elif result.status == "transient":
                attempt += 1
            else:
                raise ValueError(f"Unexpected download result status: {result.status}")
    except Exception as e:
        if result is not None:
            raise NeptuneFileDownloadError(last_status_code=result.status_code, last_content=result.content) from e
        else:
            raise NeptuneFileDownloadError() from e

    if result is not None:
        raise NeptuneFileDownloadError(
            details=f"Failed to download file after {max_tries} attempts.",
            last_status_code=result.status_code,
            last_content=result.content,
        )
    else:
        raise NeptuneFileDownloadError(details=f"Failed to download file after {max_tries} attempts.")


def create_target_path(destination: pathlib.Path, file: File) -> pathlib.Path:
    relative_target_path = pathlib.Path(".") / file.container_identifier / file.attribute_path
    if file.step is not None:
        relative_target_path = relative_target_path / f"step_{file.step:f}"
    extension = _guess_extension(file.mime_type)

    sanitized_parts = [_sanitize_path_part(part) for part in relative_target_path.parts]
    relative_target_path = pathlib.Path(*sanitized_parts)

    if extension:
        relative_target_path = relative_target_path.with_suffix(extension)

    return destination / relative_target_path


def _sanitize_path_part(part: str, max_part_length: int = 255) -> str:
    # Replace invalid characters with underscores
    part = "".join(c if c.isalnum() or c in ("_", "-") else "_" for c in part)

    if len(part) > max_part_length:
        digest = hashlib.blake2b(part.encode("utf-8"), digest_size=8).hexdigest()
        part = f"{part[:max_part_length - len(digest) - 1]}_{digest}"

    return part


def _guess_extension(mime_type: str) -> Optional[str]:
    return mimetypes.guess_extension(type=mime_type)



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from http import HTTPStatus
from typing import (
    Callable,
    Dict,
    Optional,
)

import httpx
from neptune_api import (
    AuthenticatedClient,
    Client,
)
from neptune_api.api.backend import get_client_config
from neptune_api.auth_helpers import exchange_api_key
from neptune_api.credentials import Credentials
from neptune_api.models import ClientConfig
from neptune_api.types import Response

from ..exceptions import NeptuneFailedToFetchClientConfig
from .env import (
    NEPTUNE_HTTP_REQUEST_TIMEOUT_SECONDS,
    NEPTUNE_VERIFY_SSL,
)
from .retrieval.retry import handle_errors_default


@dataclass
class TokenRefreshingURLs:
    authorization_endpoint: str
    token_endpoint: str

    @classmethod
    def from_dict(cls, data: dict) -> "TokenRefreshingURLs":
        return TokenRefreshingURLs(
            authorization_endpoint=data["authorization_endpoint"], token_endpoint=data["token_endpoint"]
        )


def _wrap_httpx_json_response(httpx_response: httpx.Response) -> Response:
    """Wrap a httpx.Response into an neptune-api Response object that is compatible
    with backoff_retry(). Use .json() as parsed content in the result."""

    return Response(
        status_code=HTTPStatus(httpx_response.status_code),
        content=httpx_response.content,
        headers=httpx_response.headers,
        parsed=httpx_response.json(),
    )


def get_config_and_token_urls(
    *, credentials: Credentials, proxies: Optional[Dict[str, str]] = None
) -> tuple[ClientConfig, TokenRefreshingURLs]:
    timeout = httpx.Timeout(NEPTUNE_HTTP_REQUEST_TIMEOUT_SECONDS.get())
    with Client(
        base_url=credentials.base_url,
        httpx_args={"mounts": proxies},
        verify_ssl=NEPTUNE_VERIFY_SSL.get(),
        timeout=timeout,
        headers={"User-Agent": _generate_user_agent()},
    ) as client:
        try:
            config_response = handle_errors_default(get_client_config.sync_detailed)(client=client)
            config = config_response.parsed

            urls_response = handle_errors_default(
                lambda: _wrap_httpx_json_response(client.get_httpx_client().get(config.security.open_id_discovery))
            )()
            token_urls = TokenRefreshingURLs.from_dict(urls_response.parsed)

            return config, token_urls
        except Exception as e:
            raise NeptuneFailedToFetchClientConfig(exception=e) from e


def create_auth_api_client(
    *,
    credentials: Credentials,
    config: ClientConfig,
    token_refreshing_urls: TokenRefreshingURLs,
    proxies: Optional[Dict[str, str]] = None,
) -> AuthenticatedClient:
    return AuthenticatedClient(
        base_url=credentials.base_url,
        credentials=credentials,
        client_id=config.security.client_id,
        token_refreshing_endpoint=token_refreshing_urls.token_endpoint,
        api_key_exchange_callback=exchange_api_key,
        verify_ssl=NEPTUNE_VERIFY_SSL.get(),
        httpx_args={"mounts": proxies, "http2": False},
        timeout=httpx.Timeout(NEPTUNE_HTTP_REQUEST_TIMEOUT_SECONDS.get()),
        headers={"User-Agent": _generate_user_agent()},
    )


_ILLEGAL_CHARS = str.maketrans({c: "_" for c in " ();/"})


def _generate_user_agent() -> str:
    import platform
    from importlib.metadata import version

    def sanitize(value: Callable[[], str]) -> str:
        try:
            result = value()
            return result.translate(_ILLEGAL_CHARS)
        except Exception:
            return "unknown"

    package_name = "neptune-query"
    package_version = sanitize(lambda: version(package_name))
    additional_metadata = {
        "neptune-api": sanitize(lambda: version("neptune-api")),
        "python": sanitize(platform.python_version),
        "os": sanitize(platform.system),
    }

    additional_metadata_str = "; ".join(f"{k}={v}" for k, v in additional_metadata.items())
    return f"{package_name}/{package_version} ({additional_metadata_str})"



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import concurrent
import contextlib
import threading
from concurrent.futures import (
    Executor,
    Future,
    ThreadPoolExecutor,
)
from typing import (
    Any,
    Callable,
    Generator,
    Iterable,
    Optional,
    ParamSpec,
    Type,
    TypeVar,
)

from .. import env

T = TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R")
OUT = tuple[set[Future[R]], Optional[R]]


def create_thread_pool_executor() -> Executor:
    max_workers = env.NEPTUNE_QUERY_MAX_WORKERS.get()
    return ThreadPoolExecutor(max_workers=max_workers)


def generate_concurrently(
    items: Generator[T, None, None],
    executor: Executor,
    downstream: Callable[[T], OUT],
) -> OUT:
    try:
        head: T = next(items)
        futures = {
            _submit_with_thread_local_propagation(executor, downstream, head),
            _submit_with_thread_local_propagation(executor, generate_concurrently, items, executor, downstream),
        }
        return futures, None
    except StopIteration:
        return set(), None


def fork_concurrently(executor: Executor, downstreams: Iterable[Callable[[], OUT]]) -> OUT:
    futures = {_submit_with_thread_local_propagation(executor, downstream) for downstream in downstreams}
    return futures, None


def return_value(item: R) -> OUT:
    return set(), item


def gather_results(output: OUT) -> Generator[R, None, None]:
    futures, value = output
    if value is not None:
        yield value
    while futures:
        done, not_done = concurrent.futures.wait(futures, return_when=concurrent.futures.FIRST_COMPLETED)
        futures = not_done
        for future in done:
            new_futures, value = future.result()
            futures.update(new_futures)
            if value is not None:
                yield value


_thread_local_storage = threading.local()


THREAD_LOCAL_PREFIX = "neptune_query_"


@contextlib.contextmanager
def use_thread_local(values: dict[str, Any]) -> Generator[None, None, None]:
    for key, value in values.items():
        setattr(_thread_local_storage, f"{THREAD_LOCAL_PREFIX}{key}", value)
    try:
        yield
    finally:
        for key in values.keys():
            attr = f"{THREAD_LOCAL_PREFIX}{key}"
            if hasattr(_thread_local_storage, attr):
                delattr(_thread_local_storage, attr)


def get_thread_local(key: str, expected_type: Type[T]) -> Optional[T]:
    value = getattr(_thread_local_storage, f"{THREAD_LOCAL_PREFIX}{key}", None)
    if value is not None and not isinstance(value, expected_type):
        raise RuntimeError(f"Expected {expected_type} for key '{key}', got {type(value)}")
    return value


def _submit_with_thread_local_propagation(
    executor: Executor, task: Callable[P, OUT], *args: P.args, **kwargs: P.kwargs
) -> Future[OUT]:
    thread_local_ctx = {
        key[len(THREAD_LOCAL_PREFIX) :]: getattr(_thread_local_storage, key)
        for key in dir(_thread_local_storage)
        if key.startswith(THREAD_LOCAL_PREFIX)
    }

    def _use_thread_local_context(downstream: Callable[P, OUT], *args: P.args, **kwargs: P.kwargs) -> OUT:
        with use_thread_local(thread_local_ctx):
            return downstream(*args, **kwargs)

    return executor.submit(_use_thread_local_context, task, *args, **kwargs)



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

from neptune_query.internal import env

NEPTUNE_LOGGER_NAME = "neptune"
DEFAULT_LOG_FORMAT = "[%(name)s] [%(levelname)s] %(message)s"
DEBUG_LOG_FORMAT = "[%(asctime)s] [%(name)s] [%(levelname)s] [%(processName)s/%(threadName)s/%(funcName)s] %(message)s"


def get_logger() -> logging.Logger:
    logger = logging.getLogger(NEPTUNE_LOGGER_NAME)
    _initialize_logger(logger)
    return logger


def _initialize_logger(logger: logging.Logger) -> None:
    if logger.hasHandlers():
        return
    level = env.NEPTUNE_LOGGER_LEVEL.get()
    logger.setLevel(level)

    handler = logging.StreamHandler()
    fmt = DEBUG_LOG_FORMAT if level == "DEBUG" else DEFAULT_LOG_FORMAT
    formatter = logging.Formatter(fmt)
    handler.setFormatter(formatter)
    logger.addHandler(handler)



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from concurrent.futures import Executor
from typing import (
    Generator,
    Iterable,
    Optional,
    Tuple,
)

from neptune_api.client import AuthenticatedClient

from .. import (
    env,
    filters,
    identifiers,
)
from ..composition import concurrency
from ..retrieval import attribute_definitions as att_defs
from ..retrieval import attribute_values as att_vals
from ..retrieval import util
from ..retrieval.attribute_filter import split_attribute_filters


def fetch_attribute_definitions(
    client: AuthenticatedClient,
    project_identifiers: Iterable[identifiers.ProjectIdentifier],
    run_identifiers: Optional[Iterable[identifiers.RunIdentifier]],
    attribute_filter: filters._BaseAttributeFilter,
    executor: Executor,
    batch_size: int = env.NEPTUNE_QUERY_ATTRIBUTE_DEFINITIONS_BATCH_SIZE.get(),
) -> Generator[util.Page[identifiers.AttributeDefinition], None, None]:
    pages_filters = _fetch_attribute_definitions(
        client, project_identifiers, run_identifiers, attribute_filter, batch_size, executor
    )

    seen_items: set[identifiers.AttributeDefinition] = set()
    for page, filter_ in pages_filters:
        new_items = [item for item in page.items if item not in seen_items]
        seen_items.update(new_items)
        yield util.Page(items=new_items)


def _fetch_attribute_definitions(
    client: AuthenticatedClient,
    project_identifiers: Iterable[identifiers.ProjectIdentifier],
    run_identifiers: Optional[Iterable[identifiers.RunIdentifier]],
    attribute_filter: filters._BaseAttributeFilter,
    batch_size: int,
    executor: Executor,
) -> Generator[tuple[util.Page[identifiers.AttributeDefinition], filters._AttributeFilter], None, None]:
    def go_fetch_single(
        filter_: filters._AttributeFilter,
    ) -> Generator[util.Page[identifiers.AttributeDefinition], None, None]:
        return att_defs.fetch_attribute_definitions_single_filter(
            client=client,
            project_identifiers=project_identifiers,
            run_identifiers=run_identifiers,
            attribute_filter=filter_,
            batch_size=batch_size,
        )

    filters_ = split_attribute_filters(attribute_filter)

    output = concurrency.generate_concurrently(
        items=(filter_ for filter_ in filters_),
        executor=executor,
        downstream=lambda filter_: concurrency.generate_concurrently(
            items=go_fetch_single(filter_),
            executor=executor,
            downstream=lambda _page: concurrency.return_value((_page, filter_)),
        ),
    )
    yield from concurrency.gather_results(output)


def fetch_attribute_values(
    client: AuthenticatedClient,
    project_identifier: identifiers.ProjectIdentifier,
    run_identifiers: Iterable[identifiers.RunIdentifier],
    attribute_filter: filters._BaseAttributeFilter,
    executor: Executor,
    batch_size: int = env.NEPTUNE_QUERY_ATTRIBUTE_DEFINITIONS_BATCH_SIZE.get(),
) -> Generator[util.Page[att_vals.AttributeValue], None, None]:
    pages_filters = _fetch_attribute_values(
        client, project_identifier, run_identifiers, attribute_filter, batch_size, executor
    )

    seen_items: set[Tuple[identifiers.RunIdentifier, identifiers.AttributeDefinition]] = set()
    for page in pages_filters:
        new_items = [item for item in page.items if (item.run_identifier, item.attribute_definition) not in seen_items]
        seen_items.update((item.run_identifier, item.attribute_definition) for item in new_items)
        yield util.Page(items=new_items)


def _fetch_attribute_values(
    client: AuthenticatedClient,
    project_identifier: identifiers.ProjectIdentifier,
    run_identifiers: Iterable[identifiers.RunIdentifier],
    attribute_filter: filters._BaseAttributeFilter,
    batch_size: int,
    executor: Executor,
) -> Generator[util.Page[att_vals.AttributeValue], None, None]:
    def go_fetch_single(filter_: filters._AttributeFilter) -> Generator[util.Page[att_vals.AttributeValue], None, None]:
        return att_vals.fetch_attribute_values(
            client=client,
            project_identifier=project_identifier,
            run_identifiers=run_identifiers,
            attribute_definitions=filter_,
            batch_size=batch_size,
        )

    filters_ = split_attribute_filters(attribute_filter)

    output = concurrency.generate_concurrently(
        items=(filter_ for filter_ in filters_),
        executor=executor,
        downstream=lambda filter_: concurrency.generate_concurrently(
            items=go_fetch_single(filter_),
            executor=executor,
            downstream=lambda _page: concurrency.return_value(_page),
        ),
    )
    yield from concurrency.gather_results(output)



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from typing import (
    Callable,
    ParamSpec,
    TypeVar,
)

from neptune_query.internal.warnings import throttled_warn
from neptune_query.warnings import ExperimentalWarning

T = ParamSpec("T")
R = TypeVar("R")


def experimental(func: Callable[T, R]) -> Callable[T, R]:
    """Decorator to mark functions as experimental.
    It will result in a warning being emitted when the function is used
    for the first time.
    """

    @functools.wraps(func)
    def wrapper(*args: T.args, **kwargs: T.kwargs) -> R:
        throttled_warn(
            ExperimentalWarning(
                f"`{func.__module__}.{func.__qualname__}` is experimental and might change or be removed "
                "in a future minor release. Use with caution in production code."
            ),
            stacklevel=3,
        )
        return func(*args, **kwargs)

    return wrapper



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = [
    "list_runs",
    "list_attributes",
    "fetch_runs_table",
    "fetch_metrics",
    "fetch_series",
    "download_files",
]

import pathlib
from typing import (
    Iterable,
    Literal,
    Optional,
    Tuple,
    Union,
)

import pandas as _pandas

from neptune_query import (
    filters,
    types,
)
from neptune_query._internal import (
    get_default_project_identifier,
    resolve_attributes_filter,
    resolve_destination_path,
    resolve_files,
    resolve_runs_filter,
    resolve_sort_by,
)
from neptune_query.exceptions import NeptuneUserError
from neptune_query.internal.composition import download_files as _download_files
from neptune_query.internal.composition import fetch_metrics as _fetch_metrics
from neptune_query.internal.composition import fetch_series as _fetch_series
from neptune_query.internal.composition import fetch_table as _fetch_table
from neptune_query.internal.composition import list_attributes as _list_attributes
from neptune_query.internal.composition import list_containers as _list_containers
from neptune_query.internal.query_metadata_context import use_query_metadata
from neptune_query.internal.retrieval import search as _search


@use_query_metadata(api_function="runs.list_runs")
def list_runs(
    *,
    project: Optional[str] = None,
    runs: Optional[Union[str, list[str], filters.Filter]] = None,
) -> list[str]:
    """Lists the IDs of runs in a Neptune project.

    Args:
        project: Path of the Neptune project, as `WorkspaceName/ProjectName`.
            If not provided, the NEPTUNE_PROJECT environment variable is used.
        runs: Filter specifying which runs to include.
            If a string is provided, it's treated as a regex pattern that the run IDs must match.
            If a list of strings is provided, it's treated as exact run IDs to match.
            To provide a more complex condition on an arbitrary attribute value, pass a Filter object.

    Examples:
        List all my runs in a specific project:
        ```
        import neptune_query.runs as nq_runs
        from neptune_query.filters import Filter


        nq_runs.list_runs(
            project="team-alpha/sandbox",
            runs=Filter.eq("sys/owner", "MyUsername"),
        )
        ```
    """
    project_identifier = get_default_project_identifier(project)
    runs_filter = resolve_runs_filter(runs)

    return _list_containers.list_containers(
        project_identifier=project_identifier,
        filter_=runs_filter,
        container_type=_search.ContainerType.RUN,
    )


@use_query_metadata(api_function="runs.list_attributes")
def list_attributes(
    *,
    project: Optional[str] = None,
    runs: Optional[Union[str, list[str], filters.Filter]] = None,
    attributes: Optional[Union[str, list[str], filters.AttributeFilter]] = None,
) -> list[str]:
    """Lists attributes in the runs of a Neptune project.

    To narrow the results, define filters for runs to search or attributes to include.

    Args:
        project: Path of the Neptune project, as `WorkspaceName/ProjectName`.
            If not provided, the NEPTUNE_PROJECT environment variable is used.
        runs: Filter specifying which runs to include.
            If a string is provided, it's treated as a regex pattern that the run IDs must match.
            If a list of strings is provided, it's treated as exact run IDs to match.
            To provide a more complex condition on an arbitrary attribute value, pass a Filter object.
        attributes: Filter specifying which attributes to include.
            If a string is provided, it's treated as a regex pattern that the attribute names must match.
            If a list of strings is provided, it's treated as exact attribute names to match.
            To provide a more complex condition, pass an AttributeFilter object.

    Examples:
        List all attributes that begin with "metrics":
        ```
        import neptune_query.runs as nq_runs


        nq_runs.list_attributes(attributes=r"^metrics")
        ```

        Search all runs of a specific project and list all attributes, except those in the
        "parameters" and "sys" namespaces:
        ```
        nq_runs.list_attributes(
            project="team-alpha/sandbox",
            attributes=r"!parameters/ & !sys/",
        )
        ```
    """
    project_identifier = get_default_project_identifier(project)
    runs_filter = resolve_runs_filter(runs)
    attributes_filter = resolve_attributes_filter(attributes)

    return _list_attributes.list_attributes(
        project_identifier=project_identifier,
        filter_=runs_filter,
        attributes=attributes_filter,
        container_type=_search.ContainerType.RUN,
    )


@use_query_metadata(api_function="runs.fetch_metrics")
def fetch_metrics(
    *,
    project: Optional[str] = None,
    runs: Union[str, list[str], filters.Filter],
    attributes: Union[str, list[str], filters.AttributeFilter],
    include_time: Optional[Literal["absolute"]] = None,
    step_range: Tuple[Optional[float], Optional[float]] = (None, None),
    lineage_to_the_root: bool = True,
    tail_limit: Optional[int] = None,
    type_suffix_in_column_names: bool = False,
    include_point_previews: bool = False,
) -> _pandas.DataFrame:
    """Fetches a table of metric values per step.

    The values are raw, without any aggregation, approximation, or interpolation.

    To narrow the results, limit the step range or the number of values from the tail end.
    You can also define filters for runs to search or attributes to include.

    Args:
        project: Path of the Neptune project, as `WorkspaceName/ProjectName`.
            If not provided, the NEPTUNE_PROJECT environment variable is used.
        runs: Filter specifying which runs to include.
            If a string is provided, it's treated as a regex pattern that the run IDs must match.
            If a list of strings is provided, it's treated as exact run IDs to match.
            To provide a more complex condition on an arbitrary attribute value, pass a Filter object.
        attributes: Filter specifying which attributes to include.
            If a string is provided, it's treated as a regex pattern that the attribute names must match.
            If a list of strings is provided, it's treated as exact attribute names to match.
            To provide a more complex condition, pass an AttributeFilter object.
        include_time: To include absolute timestamps, pass `"absolute"` as the value.
            If set, each metric column has an additional sub-column with requested timestamp values.
        step_range: Tuple specifying the range of steps to include. Can represent an open interval.
        lineage_to_the_root: If True (default), includes all points from the complete run history.
            If False, only includes points from the most recent run in the lineage.
        tail_limit: From the tail end of each series, how many points to include at most.
        type_suffix_in_column_names: If True, columns of the returned DataFrame
            are suffixed with ":<type>", e.g. "attribute1:float_series", "attribute1:string".
            If False (default), the method throws an exception if there are multiple types under one path.
        include_point_previews: If False (default), the returned results only contain committed
            points. If True, the results also include preview points and the returned DataFrame will
            have additional sub-columns with preview status (is_preview and preview_completion).

    Example:
        Fetch losses of a specific run from step 1000 onward, including incomplete points:
        ```
        import neptune_query.runs as nq_runs


        nq_runs.fetch_metrics(
            runs=["prompt-wolf-20250605132116671-2g2r1"],
            attributes=r"^loss/.*",
            step_range=(1000.0, None),
            include_point_previews=True,
        )
        ```
    """
    project_identifier = get_default_project_identifier(project)
    runs_filter = resolve_runs_filter(runs)
    attributes_filter = resolve_attributes_filter(attributes)

    return _fetch_metrics.fetch_metrics(
        project_identifier=project_identifier,
        filter_=runs_filter,
        attributes=attributes_filter,
        include_time=include_time,
        step_range=step_range,
        lineage_to_the_root=lineage_to_the_root,
        tail_limit=tail_limit,
        type_suffix_in_column_names=type_suffix_in_column_names,
        include_point_previews=include_point_previews,
        container_type=_search.ContainerType.RUN,
    )


@use_query_metadata(api_function="runs.fetch_runs_table")
def fetch_runs_table(
    *,
    project: Optional[str] = None,
    runs: Optional[Union[str, list[str], filters.Filter]] = None,
    attributes: Union[str, list[str], filters.AttributeFilter] = [],
    sort_by: Union[str, filters.Attribute] = filters.Attribute("sys/creation_time", type="datetime"),
    sort_direction: Literal["asc", "desc"] = "desc",
    limit: Optional[int] = None,
    type_suffix_in_column_names: bool = False,
) -> _pandas.DataFrame:
    """Fetches a table of run metadata, with runs as rows and attributes as columns.

    To narrow the results, define filters for runs to search or attributes to include.

    Returns a DataFrame similar to the runs table in the web app.
    For series attributes, the last logged value is returned.

    Args:
        project: Path of the Neptune project, as `WorkspaceName/ProjectName`.
            If not provided, the NEPTUNE_PROJECT environment variable is used.
        runs: Filter specifying which runs to include.
            If a string is provided, it's treated as a regex pattern that the run IDs must match.
            If a list of strings is provided, it's treated as exact run IDs to match.
            To provide a more complex condition on an arbitrary attribute value, pass a Filter object.
        attributes: Filter specifying which attributes to include.
            If a string is provided, it's treated as a regex pattern that the attribute names must match.
            If a list of strings is provided, it's treated as exact attribute names to match.
            To provide a more complex condition, pass an AttributeFilter object.
        sort_by: Name of the attribute to sort the table by.
            Alternatively, an Attribute object that specifies the attribute type.
        sort_direction: The direction to sort columns by: `"desc"` (default) or `"asc"`.
        limit: Maximum number of runs to return. By default, all runs are included.
        type_suffix_in_column_names: If True, columns of the returned DataFrame
            are suffixed with ":<type>", e.g. "attribute1:float_series", "attribute1:string".
            If False (default), the method throws an exception if there are multiple types under one path.

    Example:
        Fetch constituent runs of an experiment, with attributes matching `loss` or `configs` as columns:
        ```
        import neptune_query.runs as nq_runs
        from neptune_query.filters import Filter


        nq_runs.fetch_runs_table(
            runs=Filter.eq("sys/name", "exp-week9"),
            attributes=r"loss | configs",
        )
        ```
    """
    project_identifier = get_default_project_identifier(project)
    runs_filter = resolve_runs_filter(runs)
    attributes_filter = resolve_attributes_filter(attributes)
    resolved_sort_by = resolve_sort_by(sort_by)

    return _fetch_table.fetch_table(
        project_identifier=project_identifier,
        filter_=runs_filter,
        attributes=attributes_filter,
        sort_by=resolved_sort_by,
        sort_direction=sort_direction,
        limit=limit,
        type_suffix_in_column_names=type_suffix_in_column_names,
        container_type=_search.ContainerType.RUN,
    )


@use_query_metadata(api_function="runs.fetch_series")
def fetch_series(
    *,
    project: Optional[str] = None,
    runs: Union[str, list[str], filters.Filter],
    attributes: Union[str, list[str], filters.AttributeFilter],
    include_time: Optional[Literal["absolute"]] = None,
    step_range: Tuple[Optional[float], Optional[float]] = (None, None),
    lineage_to_the_root: bool = True,
    tail_limit: Optional[int] = None,
) -> _pandas.DataFrame:
    """Fetches a table of series values per step, for non-numerical series attributes.

    To narrow the results, define filters for runs to search or attributes to include.

    Supports series of histograms, files, and strings.

    Args:
        project: Path of the Neptune project, as `WorkspaceName/ProjectName`.
            If not provided, the NEPTUNE_PROJECT environment variable is used.
        runs: Filter specifying which runs to include.
            If a string is provided, it's treated as a regex pattern that the run IDs must match.
            If a list of strings is provided, it's treated as exact run IDs to match.
            To provide a more complex condition on an arbitrary attribute value, pass a Filter object.
        attributes: Filter specifying which attributes to include.
            If a string is provided, it's treated as a regex pattern that the attribute names must match.
            If a list of strings is provided, it's treated as exact attribute names to match.
            To provide a more complex condition, pass an AttributeFilter object.
        include_time: To include absolute timestamps, pass `"absolute"` as the value.
            If set, each metric column has an additional sub-column with requested timestamp values.
        step_range: Tuple specifying the range of steps to include. Can represent an open interval.
        lineage_to_the_root: If True (default), includes all values from the complete run history.
            If False, only includes values from the most recent run in the lineage.
        tail_limit: From the tail end of each series, how many values to include at most.

    Example:
        Fetch custom string series of a specific run from step 100 onward:
        ```
        import neptune_query.runs as nq_runs


        nq_runs.fetch_series(
            runs=["prompt-wolf-20250605132116671-2g2r1"],
            attributes=r"^messages/",
            step_range=(100.0, None),
        )
        ```
    """
    project_identifier = get_default_project_identifier(project)
    runs_filter = resolve_runs_filter(runs)
    attributes_filter = resolve_attributes_filter(attributes)

    return _fetch_series.fetch_series(
        project_identifier=project_identifier,
        filter_=runs_filter,
        attributes=attributes_filter,
        include_time=include_time,
        step_range=step_range,
        lineage_to_the_root=lineage_to_the_root,
        tail_limit=tail_limit,
        container_type=_search.ContainerType.RUN,
    )


@use_query_metadata(api_function="runs.download_files")
def download_files(
    *,
    files: Union[types.File, Iterable[types.File], _pandas.Series, _pandas.DataFrame],
    destination: Optional[Union[str, pathlib.Path]] = None,
) -> _pandas.DataFrame:
    """Downloads files from a Neptune project.

    - For file series, use `fetch_series()` to specify the content containing the files
    and pass the output to the `files` argument.
    - For individually assigned files, use `fetch_runs_table()`.

    Args:
        files: Which files to download, specified using one of the following options.
            - File object
            - Iterable of File objects
            - pandas Series containing, non-exclusively, File objects
            - pandas DataFrame containing, among other data, File objects
        destination: Directory where files will be downloaded.
            The path can be relative or absolute.
            If `None`, the current working directory (CWD) is used as the default.

    Returns:
        DataFrame mapping runs and attributes to the paths of downloaded files.

    Example:
        Specify files from a given step range of a series:
        ```
        import neptune_query.runs as nq_runs


        interesting_files = nq_runs.fetch_series(
            runs=["prompt-wolf-20250605132116671-2g2r1"],
            attributes=r"^predictions/",
            step_range=(2050.0, 2056.0),
        )

        nq_runs.download_files(files=interesting_files)
        ```
    """
    file_list = resolve_files(files)
    destination_path = resolve_destination_path(destination)

    if not all(file.run_id for file in file_list):
        raise NeptuneUserError(
            "Some files passed to nq_runs.download_files don't have associated run IDs. "
            "This is likely because you passed files not originating from the runs API. "
            "Please use files from the runs API instead by fetching them with "
            "nq_runs.fetch_series() or nq_runs.fetch_experiments_table()."
        )

    return _download_files.download_files(
        files=file_list,
        destination=destination_path,
        container_type=_search.ContainerType.RUN,
    )



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class Histogram:
    type: str
    edges: list[float]
    values: list[float]


@dataclass(frozen=True)
class File:
    project_identifier: str
    experiment_name: Optional[str]
    run_id: Optional[str]
    attribute_path: str
    step: Optional[float]
    path: str
    size_bytes: int
    mime_type: str

    def __post_init__(self) -> None:
        if not (self.experiment_name or self.run_id):
            raise ValueError("Either 'experiment_name' or 'run_id' must be set for File.")
        if self.experiment_name and self.run_id:
            raise ValueError("Only one of 'experiment_name' or 'run_id' should be set for File.")

    def __repr__(self) -> str:
        return f"File(size={_humanize_size(self.size_bytes)}, mime_type={self.mime_type})"

    @property
    def container_identifier(self) -> str:
        """Returns a label for the container based on the file's experiment name or run ID."""
        if self.experiment_name:
            return self.experiment_name
        elif self.run_id:
            return self.run_id
        else:
            # This should never happen due to __post_init__ validation
            raise ValueError("File must have either an experiment name or a run ID.")


def _humanize_size(size_bytes: int) -> str:
    """Convert bytes to a human-readable format."""
    if size_bytes < 1024:
        return f"{size_bytes} B"
    elif size_bytes < 1024**2:
        return f"{size_bytes / 1024:.2f} KB"
    elif size_bytes < 1024**3:
        return f"{size_bytes / (1024 ** 2):.2f} MB"
    else:
        return f"{size_bytes / (1024 ** 3):.2f} GB"



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools as ft
from typing import (
    Any,
    Generator,
    Iterable,
    NamedTuple,
    Optional,
    Tuple,
    Union,
)

from neptune_api.api.retrieval import get_series_values_proto
from neptune_api.client import AuthenticatedClient
from neptune_api.models import SeriesValuesRequest
from neptune_api.proto.neptune_pb.api.v1.model.series_values_pb2 import (
    ProtoPointValueDTO,
    ProtoSeriesValuesResponseDTO,
)
from neptune_api.types import UNSET

from neptune_query.internal.query_metadata_context import with_neptune_client_metadata

from ..identifiers import RunAttributeDefinition
from ..logger import get_logger
from ..retrieval import (
    retry,
    util,
)
from ..retrieval.attribute_types import (
    File,
    Histogram,
)
from .search import ContainerType

logger = get_logger()

SeriesValue = NamedTuple("SeriesValue", [("step", float), ("value", Any), ("timestamp_millis", float)])


def fetch_series_values(
    client: AuthenticatedClient,
    run_attribute_definitions: Iterable[RunAttributeDefinition],
    include_inherited: bool,
    container_type: ContainerType,
    step_range: Tuple[Union[float, None], Union[float, None]] = (None, None),
    tail_limit: Optional[int] = None,
) -> Generator[util.Page[tuple[RunAttributeDefinition, list[SeriesValue]]], None, None]:
    if not run_attribute_definitions:
        yield from []
        return

    run_attribute_definitions = list(run_attribute_definitions)
    width = len(str(len(run_attribute_definitions) - 1))
    request_id_to_run_attr_definition: dict[str, RunAttributeDefinition] = {
        f"{ix:0{width}d}": pair for ix, pair in enumerate(run_attribute_definitions)
    }

    params: dict[str, Any] = {
        "requests": [
            {
                "requestId": request_id,
                "series": {
                    "holder": {
                        "identifier": f"{run_definition.run_identifier}",
                        "type": "experiment",
                    },
                    "attribute": run_definition.attribute_definition.name,
                    "lineage": "FULL" if include_inherited else "NONE",
                    "lineageEntityType": "EXPERIMENT" if container_type == ContainerType.EXPERIMENT else "RUN",
                },
            }
            for request_id, run_definition in request_id_to_run_attr_definition.items()
        ],
        "stepRange": {"from": step_range[0], "to": step_range[1]},
        "order": "ascending" if tail_limit is None else "descending",
    }
    if tail_limit is not None:
        params["perSeriesPointsLimit"] = tail_limit

    yield from util.fetch_pages(
        client=client,
        fetch_page=_fetch_series_page,
        process_page=ft.partial(
            _process_series_page, request_id_to_run_attr_definition=request_id_to_run_attr_definition
        ),
        make_new_page_params=_make_new_series_page_params,
        params=params,
    )


def _fetch_series_page(
    client: AuthenticatedClient,
    params: dict[str, Any],
) -> ProtoSeriesValuesResponseDTO:
    logger.debug(f"Calling get_series_values_proto with params: {params}")

    body = SeriesValuesRequest.from_dict(params)
    call_api = retry.handle_errors_default(with_neptune_client_metadata(get_series_values_proto.sync_detailed))
    response = call_api(client=client, body=body, use_deprecated_string_fields=False)

    logger.debug(
        f"get_series_values_proto response status: {response.status_code}, "
        f"content length: {len(response.content) if response.content else 'no content'}"
    )
    return ProtoSeriesValuesResponseDTO.FromString(response.content)


def _process_series_page(
    data: ProtoSeriesValuesResponseDTO,
    request_id_to_run_attr_definition: dict[str, RunAttributeDefinition],
) -> util.Page[tuple[RunAttributeDefinition, list[SeriesValue]]]:
    items: dict[RunAttributeDefinition, list[SeriesValue]] = {}

    for series in data.series:
        if series.seriesValues.values:
            run_definition = request_id_to_run_attr_definition[series.requestId]
            values = [_extract_series_value(value) for value in series.seriesValues.values]
            items.setdefault(run_definition, []).extend(values)

    return util.Page(items=list(items.items()))


def _extract_series_value(value_dto: ProtoPointValueDTO) -> SeriesValue:
    value: Union[str, File, Histogram]
    obj = value_dto.object
    if obj.HasField("stringValue"):
        value = str(obj.stringValue)
    elif obj.HasField("fileRef"):
        file_ref = obj.fileRef
        value = File(
            path=file_ref.path,
            mime_type=file_ref.mimeType,
            size_bytes=file_ref.sizeBytes,
        )
    elif obj.HasField("histogram"):
        histogram = obj.histogram
        value = Histogram(
            type=str(histogram.type),
            values=list(histogram.values),
            edges=list(histogram.edges),
        )
    else:
        raise ValueError("Series has no valid value")
    return SeriesValue(value_dto.step, value, timestamp_millis=value_dto.timestamp_millis)


def _make_new_series_page_params(
    params: dict[str, Any], data: Optional[ProtoSeriesValuesResponseDTO]
) -> Optional[dict[str, Any]]:
    if data is None:
        for request in params["requests"]:
            request.pop("searchAfter", None)
        return params

    # series does not exist: series is missing from the response
    # series finished: series.HasField("searchAfter") == True, series.searchAfter.finished == True,
    #   series.searchAfter.token == 'nonempty'
    # series sent partially: series.HasField("searchAfter") == True, series.searchAfter.finished == False,
    #   series.searchAfter.token == 'nonempty'
    # series exists, but is outside the current page: series.HasField("searchAfter") == False
    # if an attribute does not exist at all, the backend will keep returning an empty searchAfter for it
    # so we stop requesting when there has been no progress, even though some requests may still be unfinished
    existing_series = {series.requestId for series in data.series}
    finished_requests = {
        series.requestId for series in data.series if series.HasField("searchAfter") and series.searchAfter.finished
    }
    updated_request_tokens = {
        series.requestId: {"searchAfter": {"finished": False, "token": series.searchAfter.token}}
        for series in data.series
        if series.HasField("searchAfter")
    }

    if not updated_request_tokens:  # no progress, stop fetching
        return None

    new_requests = [
        request | updated_request_tokens.get(request["requestId"], {"searchAfter": UNSET})
        for request in params["requests"]
        if request["requestId"] in existing_series and request["requestId"] not in finished_requests
    ]

    if not new_requests:  # all requests are finished
        return None

    params["requests"] = new_requests
    return params



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
from collections import defaultdict
from concurrent.futures import Executor
from dataclasses import dataclass
from typing import (
    Generic,
    Iterable,
    Optional,
    TypeVar,
    Union,
)

from neptune_api.client import AuthenticatedClient

from ...exceptions import AttributeTypeInferenceError
from ...warnings import AttributeWarning
from .. import (
    filters,
    identifiers,
)
from ..retrieval.attribute_types import (
    ATTRIBUTE_LITERAL,
    FILE_SERIES_AGGREGATIONS,
    FLOAT_SERIES_AGGREGATIONS,
    HISTOGRAM_SERIES_AGGREGATIONS,
    STRING_SERIES_AGGREGATIONS,
)
from ..warnings import throttled_warn
from .attributes import fetch_attribute_definitions

T = TypeVar("T", bound=Union[filters._Filter, filters._Attribute, None])


@dataclass
class AttributeInferenceState:
    original_attribute: filters._Attribute
    attribute: filters._Attribute
    inferred_type: Optional[str] = None
    success_details: Optional[str] = None
    error: Optional[str] = None
    warning_text: Optional[str] = None

    def is_finalized(self) -> bool:
        return self.inferred_type is not None or self.error is not None

    def is_inferred(self) -> bool:
        return self.inferred_type is not None

    def set_success(self, inferred_type: ATTRIBUTE_LITERAL, success_details: Optional[str] = None) -> None:
        self.inferred_type = inferred_type
        self.success_details = success_details
        self.attribute.type = inferred_type

    def set_error(self, error: str) -> None:
        self.error = error

    def set_fallback(
        self,
        inferred_type: ATTRIBUTE_LITERAL,
        success_details: Optional[str] = None,
        warning_text: Optional[str] = None,
    ) -> None:
        self.inferred_type = inferred_type
        self.success_details = success_details
        self.attribute.type = inferred_type
        self.error = None
        self.warning_text = warning_text


@dataclass
class InferenceState(Generic[T]):
    attributes: list[AttributeInferenceState]
    result: T

    @staticmethod
    def from_attribute(attribute: filters._Attribute) -> "InferenceState[filters._Attribute]":
        attribute_copy = copy.deepcopy(attribute)

        attribute_states = [
            AttributeInferenceState(
                original_attribute=copy.copy(attribute_copy),
                attribute=attribute_copy,
                inferred_type=attribute_copy.type,
                success_details="Type provided" if attribute_copy.type is not None else None,
            )
        ]
        return InferenceState(attributes=attribute_states, result=attribute_copy)

    @staticmethod
    def from_filter(filter_: filters._Filter) -> "InferenceState[Optional[filters._Filter]]":
        def _walk_attributes(experiment_filter: filters._Filter) -> Iterable[filters._Attribute]:
            if isinstance(experiment_filter, filters._AttributeValuePredicate):
                yield experiment_filter.attribute
            elif isinstance(experiment_filter, filters._AttributePredicate):
                yield experiment_filter.attribute
            elif isinstance(experiment_filter, filters._AssociativeOperator):
                for child in experiment_filter.filters:
                    yield from _walk_attributes(child)
            elif isinstance(experiment_filter, filters._PrefixOperator):
                yield from _walk_attributes(experiment_filter.filter_)
            else:
                raise RuntimeError(f"Unexpected filter type: {type(experiment_filter)}")

        filter_copy = copy.deepcopy(filter_)

        attributes = _walk_attributes(filter_copy)
        attribute_states = [
            AttributeInferenceState(
                original_attribute=copy.copy(attr),
                attribute=attr,
                inferred_type=attr.type,
                success_details="Type provided" if attr.type is not None else None,
            )
            for attr in attributes
        ]
        return InferenceState(attributes=attribute_states, result=filter_copy)

    def incomplete_attributes(self) -> list[AttributeInferenceState]:
        return [attr_state for attr_state in self.attributes if not attr_state.is_finalized()]

    def is_complete(self) -> bool:
        return all(attr_state.is_finalized() for attr_state in self.attributes)

    def raise_if_incomplete(self) -> None:
        uninferred_attributes = [attr_state for attr_state in self.attributes if not attr_state.is_inferred()]
        if uninferred_attributes:
            attribute_names = [attr_state.original_attribute.name for attr_state in uninferred_attributes]

            details = []
            for attr_state in uninferred_attributes:
                if attr_state.error:
                    details.append(f"{attr_state.original_attribute.name}: {attr_state.error}")
                else:
                    details.append(f"{attr_state.original_attribute.name}: could not find the attribute")

            raise AttributeTypeInferenceError(attribute_names=attribute_names, details=details)

    def get_result_or_raise(self) -> T:
        self.raise_if_incomplete()
        return self.result

    def emit_warnings(self) -> None:
        for attr_state in self.attributes:
            if attr_state.warning_text:
                throttled_warn(
                    AttributeWarning(f"Attribute '{attr_state.original_attribute.name}': {attr_state.warning_text}"),
                    stacklevel=3,
                )


def infer_attribute_types_in_filter(
    client: AuthenticatedClient,
    project_identifier: identifiers.ProjectIdentifier,
    filter_: Optional[filters._Filter],
    fetch_attribute_definitions_executor: Executor,
) -> InferenceState[Optional[filters._Filter]]:
    if filter_ is None:
        return InferenceState(attributes=[], result=None)

    state = InferenceState.from_filter(filter_)
    if state.is_complete():
        return state

    _infer_attribute_types_locally(inference_state=state)
    if state.is_complete():
        return state

    _infer_attribute_types_from_api(
        client=client,
        project_identifier=project_identifier,
        fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
        inference_state=state,
    )

    _fill_unknown_types_as_string(state)
    return state


def infer_attribute_types_in_sort_by(
    client: AuthenticatedClient,
    project_identifier: identifiers.ProjectIdentifier,
    sort_by: filters._Attribute,
    fetch_attribute_definitions_executor: Executor,
) -> InferenceState[filters._Attribute]:
    state = InferenceState.from_attribute(sort_by)
    if state.is_complete():
        return state

    _infer_attribute_types_locally(inference_state=state)
    if state.is_complete():
        return state

    _infer_attribute_types_from_api(
        client=client,
        project_identifier=project_identifier,
        fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
        inference_state=state,
    )

    _fill_unknown_types_as_string(state)
    _fill_conflicting_types_as_string(state)
    return state


_KNOWN_SYS_ATTRIBUTES: dict[str, ATTRIBUTE_LITERAL] = {
    "sys/archived": "bool",
    "sys/creation_time": "datetime",
    "sys/custom_run_id": "string",
    "sys/description": "string",
    "sys/diagnostics/attributes/bool_count": "int",
    "sys/diagnostics/attributes/file_ref_count": "int",
    "sys/diagnostics/attributes/file_ref_series_count": "int",
    "sys/diagnostics/attributes/float_count": "int",
    "sys/diagnostics/attributes/float_series_count": "int",
    "sys/diagnostics/attributes/histogram_count": "int",
    "sys/diagnostics/attributes/histogram_series_count": "int",
    "sys/diagnostics/attributes/int_count": "int",
    "sys/diagnostics/attributes/string_count": "int",
    "sys/diagnostics/attributes/string_series_count": "int",
    "sys/diagnostics/attributes/string_set_count": "int",
    "sys/diagnostics/attributes/total_count": "int",
    "sys/diagnostics/attributes/total_series_datapoints": "int",
    "sys/diagnostics/project_uuid": "string",
    "sys/diagnostics/run_uuid": "string",
    "sys/experiment/is_head": "bool",
    "sys/experiment/name": "string",
    "sys/experiment/running_time_seconds": "float",
    "sys/failed": "bool",
    "sys/family": "string",
    "sys/forking/depth": "int",
    "sys/group_tags": "string_set",
    "sys/id": "string",
    "sys/modification_time": "datetime",
    "sys/name": "string",
    "sys/owner": "string",
    "sys/ping_time": "datetime",
    "sys/relative_creation_time_ms": "int",
    "sys/running_time_seconds": "float",
    "sys/size": "int",
    "sys/tags": "string_set",
    "sys/trashed": "bool",
}


def _infer_attribute_types_locally(
    inference_state: InferenceState,
) -> None:
    for state in inference_state.incomplete_attributes():
        attribute = state.attribute
        if attribute.name in _KNOWN_SYS_ATTRIBUTES:
            inferred_type = _KNOWN_SYS_ATTRIBUTES[attribute.name]
            state.set_success(
                inferred_type=inferred_type,
                success_details="Inferred as a known system attribute",
            )

    # TODO: this doesn't have a lot of sense in neptune-query, as we don't have aggregations beyond last anymore
    for state in inference_state.incomplete_attributes():
        attribute = state.attribute
        matches: list[ATTRIBUTE_LITERAL] = []
        if all(agg in FLOAT_SERIES_AGGREGATIONS for agg in attribute.aggregation or []):
            matches.append("float_series")
        if all(agg in STRING_SERIES_AGGREGATIONS for agg in attribute.aggregation or []):
            matches.append("string_series")
        if all(agg in FILE_SERIES_AGGREGATIONS for agg in attribute.aggregation or []):
            matches.append("file_series")
        if all(agg in HISTOGRAM_SERIES_AGGREGATIONS for agg in attribute.aggregation or []):
            matches.append("histogram_series")
        if len(matches) == 1:
            state.set_success(inferred_type=matches[0], success_details="Inferred from aggregation")


def _infer_attribute_types_from_api(
    client: AuthenticatedClient,
    project_identifier: identifiers.ProjectIdentifier,
    fetch_attribute_definitions_executor: Executor,
    inference_state: InferenceState,
) -> None:
    attribute_states = inference_state.incomplete_attributes()
    attributes = [state.attribute for state in attribute_states]
    attribute_filter_by_name = filters._AttributeFilter(name_eq=list({attr.name for attr in attributes}))

    output = fetch_attribute_definitions(
        client=client,
        project_identifiers=[project_identifier],
        run_identifiers=None,
        attribute_filter=attribute_filter_by_name,
        executor=fetch_attribute_definitions_executor,
    )

    inferred_attributes = defaultdict(set)

    for page in output:
        for attr_def in page.items:
            for state in attribute_states:
                if state.attribute.name == attr_def.name:
                    inferred_attributes[state.attribute.name].add(attr_def.type)

    for state in attribute_states:
        types = inferred_attributes.get(state.attribute.name, set())
        if len(types) == 1:
            state.set_success(
                inferred_type=list(types)[0],  # type: ignore
                success_details="Inferred from neptune api",
            )

        if len(types) > 1:
            state.set_error(
                error=f"Neptune found the attribute name in multiple runs and experiments across the project "
                f"with conflicting types: {', '.join(types)}"
            )


def _fill_unknown_types_as_string(state: InferenceState) -> None:
    """
    Fills the types of attributes that are not inferred with "string".
    """
    for attr_state in state.incomplete_attributes():
        if attr_state.inferred_type is None:
            attr_state.set_fallback(
                inferred_type="string",
                success_details="Defaulting to string type for unknown attributes",
                warning_text="Attribute doesn't exist in the project",
            )


def _fill_conflicting_types_as_string(state: InferenceState) -> None:
    """
    Fills the types of attributes that have conflicting types with "string".
    """
    for attr_state in state.attributes:
        if attr_state.inferred_type is None and attr_state.error:
            attr_state.set_fallback(
                inferred_type="string",
                success_details="Defaulting to string type for attributes with conflicting types",
                warning_text=attr_state.error,
            )



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dataclasses import dataclass
from typing import NewType

ProjectIdentifier = NewType("ProjectIdentifier", str)  # e.g. "team/john.doe"
SysId = NewType("SysId", str)  # e.g. "KEY-1234"
SysName = NewType("SysName", str)  # e.g. "pye2e-fetcher-test-internal-attribute"
CustomRunId = NewType("CustomRunId", str)  # an uuid


@dataclass(frozen=True)
class RunIdentifier:
    project_identifier: ProjectIdentifier
    sys_id: SysId

    def __str__(self) -> str:
        return f"{self.project_identifier}/{self.sys_id}"


@dataclass(frozen=True)
class AttributeDefinition:
    name: str
    type: str


@dataclass(frozen=True)
class RunAttributeDefinition:
    run_identifier: RunIdentifier
    attribute_definition: AttributeDefinition
